from pathlib import PurePath

import pytest
from PyQt6.QtCore import QDateTime, QItemSelectionModel, Qt
from PyQt6.QtWidgets import QMenu

import vorta.borg
import vorta.utils
import vorta.views.archive_tab
from vorta.store.models import ArchiveModel
from vorta.views.diff_result import (
    ChangeType,
    DiffData,
    DiffResultDialog,
    DiffTree,
    FileType,
    parse_diff_json,
    parse_diff_lines,
)
from vorta.views.partials.treemodel import FileTreeModel


def setup_diff_result_window(qtbot, mocker, tab, borg_json_output, json_mock_file="diff_archives"):
    """Sets up the diff result window."""
    stdout, stderr = borg_json_output(json_mock_file)
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    compat = vorta.utils.borg_compat

    def check(feature_name):
        if feature_name == 'DIFF_JSON_LINES':
            return False
        return vorta.utils.BorgCompatibility.check(compat, feature_name)

    mocker.patch.object(vorta.utils.borg_compat, 'check', check)

    selection_model: QItemSelectionModel = tab.archiveTable.selectionModel()
    model = tab.archiveTable.model()

    flags = QItemSelectionModel.SelectionFlag.Rows
    flags |= QItemSelectionModel.SelectionFlag.Select

    selection_model.select(model.index(0, 0), flags)
    selection_model.select(model.index(1, 0), flags)

    tab.diff_action()

    qtbot.waitUntil(lambda: hasattr(tab, '_resultwindow'), **pytest._wait_defaults)
    assert hasattr(tab, '_resultwindow')


@pytest.mark.parametrize(
    'json_mock_file, folder_root', [('diff_archives', 'test'), ('diff_archives_dict_issue', 'Users')]
)
def test_archive_diff(qapp, qtbot, mocker, borg_json_output, json_mock_file, folder_root, archive_env):
    """Tests basic functionality of archive diff."""
    main, tab = archive_env
    setup_diff_result_window(qtbot, mocker, tab, borg_json_output, json_mock_file)

    model = tab._resultwindow.treeView.model().sourceModel()
    assert model.root.children[0].subpath == folder_root
    assert tab._resultwindow.archiveNameLabel_1.text() == 'test-archive'
    tab._resultwindow.accept()


def test_diff_item_copy(qapp, qtbot, mocker, borg_json_output, archive_env):
    """Tests copy action by row selection and when passed an index."""
    main, tab = archive_env
    setup_diff_result_window(qtbot, mocker, tab, borg_json_output)

    # mock the clipboard to ensure no changes are made to it during testing
    mocker.patch.object(qapp.clipboard(), "setMimeData")
    clipboard_spy = mocker.spy(qapp.clipboard(), "setMimeData")

    # test 'diff_item_copy()' by passing it an item to copy
    index = tab._resultwindow.treeView.model().index(0, 0)
    assert index is not None
    tab._resultwindow.diff_item_copy(index)
    clipboard_data = clipboard_spy.call_args[0][0]
    assert clipboard_data.hasText()
    assert clipboard_data.text() == "/test"

    clipboard_spy.reset_mock()

    # test 'diff_item_copy()' by selecting a row to copy
    flags = QItemSelectionModel.SelectionFlag.Rows
    flags |= QItemSelectionModel.SelectionFlag.Select
    tab._resultwindow.treeView.selectionModel().select(tab._resultwindow.treeView.model().index(0, 0), flags)
    tab._resultwindow.diff_item_copy()
    clipboard_data = clipboard_spy.call_args[0][0]
    assert clipboard_data.hasText()
    assert clipboard_data.text() == "/test"


def test_treeview_context_menu(qapp, qtbot, mocker, borg_json_output, archive_env):
    """Tests the diff result window context menu for expected actions."""
    main, tab = archive_env
    setup_diff_result_window(qtbot, mocker, tab, borg_json_output)

    # Load the context menu at the first result in window
    pos = tab._resultwindow.treeView.visualRect(tab._resultwindow.treeView.model().index(0, 0)).center()
    tab._resultwindow.treeview_context_menu(pos)
    qtbot.waitUntil(lambda: tab._resultwindow.findChild(QMenu) is not None, **pytest._wait_defaults)
    context_menu = tab._resultwindow.findChild(QMenu)
    assert context_menu is not None

    # assert the actions are available in the context menu
    expected_actions = ['Copy', 'Expand recursively']
    for action in expected_actions:
        assert any(menu_actions.text() == action for menu_actions in context_menu.actions())


@pytest.mark.parametrize(
    'line, expected',
    [
        (
            'changed link        some/changed/link',
            ('some/changed/link', FileType.LINK, ChangeType.CHANGED_LINK, 0, 0, None, None, None, None, None),
        ),
        (
            ' +77.8 kB  -77.8 kB some/changed/file',
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODIFIED,
                2 * 77800,
                0,
                None,
                None,
                None,
                None,
                (77800, 77800),
            ),
        ),
        (
            ' +77.8 kB  -77.8 kB [-rw-rw-rw- -> -rw-r--r--] some/changed/file',
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODIFIED,
                2 * 77800,
                0,
                ('-rw-rw-rw-', '-rw-r--r--'),
                None,
                None,
                None,
                (77800, 77800),
            ),
        ),
        (
            '[-rw-rw-rw- -> -rw-r--r--] some/changed/file',
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODE,
                0,
                0,
                ('-rw-rw-rw-', '-rw-r--r--'),
                None,
                None,
                None,
                None,
            ),
        ),
        (
            'added directory    some/changed/dir',
            ('some/changed/dir', FileType.DIRECTORY, ChangeType.ADDED, 0, 0, None, None, None, None, None),
        ),
        (
            'removed directory  some/changed/dir',
            ('some/changed/dir', FileType.DIRECTORY, ChangeType.REMOVED_DIR, 0, 0, None, None, None, None, None),
        ),
        # Example from https://github.com/borgbase/vorta/issues/521
        (
            '[user:user -> nfsnobody:nfsnobody] home/user/arrays/test.txt',
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                0,
                0,
                None,
                ('user', 'user', 'nfsnobody', 'nfsnobody'),
                None,
                None,
                None,
            ),
        ),
        # Very short owner change, to check stripping whitespace from file path
        (
            '[a:a -> b:b]       home/user/arrays/test.txt',
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                0,
                0,
                None,
                ('a', 'a', 'b', 'b'),
                None,
                None,
                None,
            ),
        ),
        # All file-related changes in one test
        (
            ' +77.8 kB  -800 B [user:user -> nfsnobody:nfsnobody] [-rw-rw-rw- -> -rw-r--r--] home/user/arrays/test.txt',
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                77800 + 800,
                77000,
                ('-rw-rw-rw-', '-rw-r--r--'),
                ('user', 'user', 'nfsnobody', 'nfsnobody'),
                None,
                None,
                (77800, 800),
            ),
        ),
    ],
)
def test_archive_diff_parser(line, expected):
    model = DiffTree()
    model.setMode(model.DisplayMode.FLAT)
    parse_diff_lines([line], model)

    assert model.rowCount() == 1
    item = model.index(0, 0).internalPointer()

    assert item.path == PurePath(expected[0]).parts
    assert item.data == DiffData(*expected[1:])


@pytest.mark.parametrize(
    'line, expected',
    [
        (
            {'path': 'some/changed/link', 'changes': [{'type': 'changed link'}]},
            ('some/changed/link', FileType.LINK, ChangeType.CHANGED_LINK, 0, 0, None, None, None, None, None),
        ),
        (
            {'path': 'some/changed/file', 'changes': [{'type': 'modified', 'added': 77800, 'removed': 77800}]},
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODIFIED,
                2 * 77800,
                0,
                None,
                None,
                None,
                None,
                (77800, 77800),
            ),
        ),
        (
            {
                'path': 'some/changed/file',
                'changes': [
                    {'type': 'modified', 'added': 77800, 'removed': 800},
                    {'type': 'mode', 'old_mode': '-rw-rw-rw-', 'new_mode': '-rw-r--r--'},
                ],
            },
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODIFIED,
                77800 + 800,
                77000,
                ('-rw-rw-rw-', '-rw-r--r--'),
                None,
                None,
                None,
                (77800, 800),
            ),
        ),
        (
            {
                'path': 'some/changed/file',
                'changes': [{'type': 'mode', 'old_mode': '-rw-rw-rw-', 'new_mode': '-rw-r--r--'}],
            },
            (
                'some/changed/file',
                FileType.FILE,
                ChangeType.MODE,
                0,
                0,
                ('-rw-rw-rw-', '-rw-r--r--'),
                None,
                None,
                None,
                None,
            ),
        ),
        (
            {'path': 'some/changed/dir', 'changes': [{'type': 'added directory'}]},
            ('some/changed/dir', FileType.DIRECTORY, ChangeType.ADDED, 0, 0, None, None, None, None, None),
        ),
        (
            {'path': 'some/changed/dir', 'changes': [{'type': 'removed directory'}]},
            ('some/changed/dir', FileType.DIRECTORY, ChangeType.REMOVED_DIR, 0, 0, None, None, None, None, None),
        ),
        # Example from https://github.com/borgbase/vorta/issues/521
        (
            {
                'path': 'home/user/arrays/test.txt',
                'changes': [
                    {
                        'type': 'owner',
                        'old_user': 'user',
                        'new_user': 'nfsnobody',
                        'old_group': 'user',
                        'new_group': 'nfsnobody',
                    }
                ],
            },
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                0,
                0,
                None,
                ('user', 'user', 'nfsnobody', 'nfsnobody'),
                None,
                None,
                None,
            ),
        ),
        # Very short owner change, to check stripping whitespace from file path
        (
            {
                'path': 'home/user/arrays/test.txt',
                'changes': [{'type': 'owner', 'old_user': 'a', 'new_user': 'b', 'old_group': 'a', 'new_group': 'b'}],
            },
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                0,
                0,
                None,
                ('a', 'a', 'b', 'b'),
                None,
                None,
                None,
            ),
        ),
        # Short ctime change
        (
            {
                'path': 'home/user/arrays',
                'changes': [
                    {
                        'new_ctime': '2023-04-01T17:23:14.104630',
                        'old_ctime': '2023-03-03T23:40:17.073948',
                        'type': 'ctime',
                    }
                ],
            },
            (
                'home/user/arrays',
                FileType.FILE,
                ChangeType.MODIFIED,
                0,
                0,
                None,
                None,
                (
                    QDateTime.fromString('2023-03-03T23:40:17.073948', Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString('2023-04-01T17:23:14.104630', Qt.DateFormat.ISODateWithMs),
                ),
                None,
                None,
            ),
        ),
        # Short mtime change
        (
            {
                'path': 'home/user/arrays',
                'changes': [
                    {
                        'new_mtime': '2023-04-01T17:23:14.104630',
                        'old_mtime': '2023-03-03T23:40:17.073948',
                        'type': 'mtime',
                    }
                ],
            },
            (
                'home/user/arrays',
                FileType.FILE,
                ChangeType.MODIFIED,
                0,
                0,
                None,
                None,
                None,
                (
                    QDateTime.fromString('2023-03-03T23:40:17.073948', Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString('2023-04-01T17:23:14.104630', Qt.DateFormat.ISODateWithMs),
                ),
                None,
            ),
        ),
        # All file-related changes in one test
        (
            {
                'path': 'home/user/arrays/test.txt',
                'changes': [
                    {'type': 'modified', 'added': 77800, 'removed': 77800},
                    {'type': 'mode', 'old_mode': '-rw-rw-rw-', 'new_mode': '-rw-r--r--'},
                    {
                        'type': 'owner',
                        'old_user': 'user',
                        'new_user': 'nfsnobody',
                        'old_group': 'user',
                        'new_group': 'nfsnobody',
                    },
                    {
                        'new_ctime': '2023-04-01T17:23:14.104630',
                        'old_ctime': '2023-03-03T23:40:17.073948',
                        'type': 'ctime',
                    },
                    {
                        'new_mtime': '2023-04-01T17:15:50.290565',
                        'old_mtime': '2023-03-05T00:24:00.359045',
                        'type': 'mtime',
                    },
                ],
            },
            (
                'home/user/arrays/test.txt',
                FileType.FILE,
                ChangeType.OWNER,
                2 * 77800,
                0,
                ('-rw-rw-rw-', '-rw-r--r--'),
                ('user', 'user', 'nfsnobody', 'nfsnobody'),
                (
                    QDateTime.fromString('2023-03-03T23:40:17.073948', Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString('2023-04-01T17:23:14.104630', Qt.DateFormat.ISODateWithMs),
                ),
                (
                    QDateTime.fromString('2023-03-05T00:24:00.359045', Qt.DateFormat.ISODateWithMs),
                    QDateTime.fromString('2023-04-01T17:15:50.290565', Qt.DateFormat.ISODateWithMs),
                ),
                (77800, 77800),
            ),
        ),
    ],
)
def test_archive_diff_json_parser(line, expected):
    model = DiffTree()
    model.setMode(model.DisplayMode.FLAT)
    parse_diff_json([line], model)

    assert model.rowCount() == 1
    item = model.index(0, 0).internalPointer()

    assert item.path == PurePath(expected[0]).parts
    assert item.data == DiffData(*expected[1:])


@pytest.mark.parametrize(
    "selection, expected_mode, expected_bCollapseAllEnabled",
    [
        (0, FileTreeModel.DisplayMode.TREE, True),
        (1, FileTreeModel.DisplayMode.SIMPLIFIED_TREE, True),
        (2, FileTreeModel.DisplayMode.FLAT, False),
    ],
)
def test_change_display_mode(selection: int, expected_mode, expected_bCollapseAllEnabled):
    dialog = DiffResultDialog(ArchiveModel(), ArchiveModel(), DiffTree())
    dialog.change_display_mode(selection)

    assert dialog.model.mode == expected_mode
    assert dialog.bCollapseAll.isEnabled() == expected_bCollapseAllEnabled



"""
This file contains tests that directly call borg commands and verify the exit code.
"""

from pathlib import Path

import pytest

import vorta.borg
import vorta.store.models
from vorta.borg.info_archive import BorgInfoArchiveJob
from vorta.borg.info_repo import BorgInfoRepoJob
from vorta.borg.prune import BorgPruneJob


def test_borg_prune(qapp, qtbot):
    """This test runs borg prune on a test repo directly without UI"""
    params = BorgPruneJob.prepare(vorta.store.models.BackupProfileModel.select().first())
    thread = BorgPruneJob(params['cmd'], params, qapp)

    with qtbot.waitSignal(thread.result, **pytest._wait_defaults) as blocker:
        blocker.connect(thread.updated)
        thread.run()

    assert blocker.args[0]['returncode'] == 0


def test_borg_repo_info(qapp, qtbot, tmpdir):
    """This test runs borg info on a test repo directly without UI"""
    repo_info = {
        'repo_url': str(Path(tmpdir).parent / 'repo0'),
        'repo_name': 'repo0',
        'extra_borg_arguments': '',
        'ssh_key': '',
        'password': '',
    }

    params = BorgInfoRepoJob.prepare(repo_info)
    thread = BorgInfoRepoJob(params['cmd'], params, qapp)

    with qtbot.waitSignal(thread.result, **pytest._wait_defaults) as blocker:
        blocker.connect(thread.result)
        thread.run()

    assert blocker.args[0]['returncode'] == 0


def test_borg_archive_info(qapp, qtbot, archive_env):
    """Check that archive info command works"""
    params = BorgInfoArchiveJob.prepare(vorta.store.models.BackupProfileModel.select().first(), "test-archive1")
    thread = BorgInfoArchiveJob(params['cmd'], params, qapp)

    with qtbot.waitSignal(thread.result, **pytest._wait_defaults) as blocker:
        blocker.connect(thread.result)
        thread.run()

    assert blocker.args[0]['returncode'] == 0



from unittest.mock import patch

import pytest
from PyQt6.QtCore import QVariant

from vorta.keyring.kwallet import KWalletNotAvailableException, VortaKWallet5Keyring


@pytest.fixture
def kwallet_keyring():
    with patch('vorta.keyring.kwallet.QtDBus.QDBusInterface') as MockInterface:
        mock_iface = MockInterface.return_value
        mock_iface.isValid.return_value = True

        with patch.object(VortaKWallet5Keyring, 'get_result') as mock_get_result:
            mock_get_result.side_effect = lambda method, args=[]: (True if method == "isEnabled" else None)

            mock_iface.callWithArgumentList.return_value.arguments.return_value = [1]
            yield VortaKWallet5Keyring()


@patch('vorta.keyring.kwallet.QtDBus.QDBusInterface')
def test_init_valid(mock_iface):
    mock_iface.return_value.isValid.return_value = True
    with patch.object(VortaKWallet5Keyring, 'get_result', return_value=True):
        keyring = VortaKWallet5Keyring()
        assert keyring.iface.isValid()


@patch('vorta.keyring.kwallet.QtDBus.QDBusInterface')
def test_init_invalid(mock_iface):
    mock_iface.return_value.isValid.return_value = False

    with pytest.raises(KWalletNotAvailableException):
        VortaKWallet5Keyring()


def test_set_password(kwallet_keyring):
    with patch.object(kwallet_keyring, 'get_result', return_value=None) as mock_get_result:
        kwallet_keyring.set_password('test_service', 'test_repo', 'test_password')
        mock_get_result.assert_called_once_with(
            "writePassword",
            args=[kwallet_keyring.handle, kwallet_keyring.folder_name, 'test_repo', 'test_password', 'test_service'],
        )


@patch('vorta.keyring.kwallet.QInputDialog.getText', return_value=('', False))
def test_get_password(mock_dialog, kwallet_keyring):
    wId = QVariant(0)

    with patch.object(kwallet_keyring, 'get_result') as mock_get_result:
        mock_get_result.side_effect = [
            'test_wallet',  # networkWallet
            42,  # open (wallet handle)
            True,  # hasEntry
            'test_password',  # readPassword
        ]

        password = kwallet_keyring.get_password('test_service', 'test_repo')

        # Debug assertions to ensure proper flow
        mock_get_result.assert_any_call("networkWallet")
        mock_get_result.assert_any_call("open", args=['test_wallet', wId, 'vorta-repo'])
        mock_get_result.assert_any_call(
            "hasEntry", args=[kwallet_keyring.handle, kwallet_keyring.folder_name, 'test_repo', 'test_service']
        )
        mock_get_result.assert_any_call(
            "readPassword", args=[kwallet_keyring.handle, kwallet_keyring.folder_name, 'test_repo', 'test_service']
        )

        assert password == 'test_password'


@patch('vorta.keyring.kwallet.QInputDialog.getText', return_value=('', False))
def test_get_password_not_found(mock_dialog, kwallet_keyring):
    kwallet_keyring.iface.callWithArgumentList.return_value.arguments.return_value = [False]

    password = kwallet_keyring.get_password('test_service', 'test_repo')
    assert password is None


@patch('vorta.keyring.kwallet.QInputDialog.getText', return_value=('', False))
def test_try_unlock(mock_dialog, kwallet_keyring):
    with patch.object(kwallet_keyring, 'get_result') as mock_get_result:
        mock_get_result.side_effect = [
            'test_wallet',  # networkWallet
            42,  # open (wallet handle)
        ]
        kwallet_keyring.try_unlock()
        assert kwallet_keyring.handle == 42



import os
import socket
import sys
from unittest.mock import Mock

import pytest
from peewee import SqliteDatabase

import vorta
import vorta.application
import vorta.borg.jobs_manager


def pytest_configure(config):
    pytest._wait_defaults = {'timeout': 20000}
    os.environ['LANG'] = 'en'  # Ensure we test an English UI

    # Disable pytest-qt's processEvents() calls on macOS to prevent hangs.
    # See: https://github.com/pytest-dev/pytest-qt/issues/223
    if sys.platform == 'darwin':
        try:
            import pytestqt.plugin

            pytestqt.plugin._process_events = lambda: None
        except (ImportError, AttributeError):
            pass

    # Mock D-Bus system bus to prevent hangs in CI (scheduler.py, network_manager.py)
    try:
        from PyQt6 import QtDBus

        _original_system_bus = QtDBus.QDBusConnection.systemBus

        def _mock_system_bus():
            mock_bus = Mock()
            mock_bus.isConnected.return_value = False
            return mock_bus

        QtDBus.QDBusConnection.systemBus = staticmethod(_mock_system_bus)
    except ImportError:
        pass

    # Mock DNS lookups to prevent timeouts in CI (utils._getfqdn)
    _original_getaddrinfo = socket.getaddrinfo
    socket.getaddrinfo = lambda *args, **kwargs: []

    # Mock WiFi enumeration to prevent hangs on headless CI (utils.get_sorted_wifis)
    import vorta.utils

    _original_get_network_status_monitor = vorta.utils.get_network_status_monitor

    def _mock_get_network_status_monitor():
        mock_monitor = Mock()
        mock_monitor.get_known_wifis.return_value = []
        return mock_monitor

    vorta.utils.get_network_status_monitor = _mock_get_network_status_monitor


@pytest.fixture(scope='session')
def qapp(tmpdir_factory):
    # DB is required to init QApplication. New DB used for every test.
    tmp_db = tmpdir_factory.mktemp('Vorta').join('settings.sqlite')
    mock_db = SqliteDatabase(str(tmp_db))
    vorta.store.connection.init_db(mock_db)

    # Needs to be disabled before calling VortaApp()
    if sys.platform == 'darwin':
        cfg = vorta.store.models.SettingsModel.get(key='check_full_disk_access')
        cfg.value = False
        cfg.save()

    # Force use of DB keyring instead of system keyring to avoid keychain prompts during tests
    keyring_setting = vorta.store.models.SettingsModel.get(key='use_system_keyring')
    keyring_setting.value = False
    keyring_setting.save()

    from vorta.application import VortaApp

    qapp = VortaApp([])  # Only init QApplication once to avoid segfaults while testing.

    yield qapp
    mock_db.close()
    qapp.quit()



"""
Test backup creation
"""

import pytest
from PyQt6 import QtCore

from vorta.store.models import ArchiveModel, EventLogModel


def test_create(qapp, qtbot, archive_env):
    """Test for manual archive creation"""
    main, tab = archive_env

    qtbot.mouseClick(main.createStartBtn, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: 'Backup finished.' in main.progressText.text(), **pytest._wait_defaults)
    qtbot.waitUntil(lambda: main.createStartBtn.isEnabled(), **pytest._wait_defaults)

    assert EventLogModel.select().count() == 2
    assert ArchiveModel.select().count() == 7
    assert main.createStartBtn.isEnabled()
    assert main.archiveTab.archiveTable.rowCount() == 7
    assert main.scheduleTab.logPage.logPage.rowCount() == 2



import os
import sys
from pathlib import Path
from unittest.mock import Mock

import pytest
from PyQt6 import QtCore
from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QCheckBox, QFormLayout, QMessageBox

import vorta.store.models
from vorta.store.models import SettingsModel


def test_toggle_all_settings(qapp, qtbot):
    """Toggle each setting twice as a basic sanity test to ensure app does crash."""
    groups = (
        SettingsModel.select(SettingsModel.group)
        .distinct(True)
        .where(SettingsModel.group != '')
        .order_by(SettingsModel.group.asc())
    )

    settings = [
        setting
        for group in groups
        for setting in SettingsModel.select().where(
            SettingsModel.type == 'checkbox', SettingsModel.group == group.group
        )
    ]

    for setting in settings:
        for _ in range(2):
            _click_toggle_setting(setting.label, qapp, qtbot)


@pytest.mark.skipif(sys.platform != "linux", reason="testing autostart path for Linux only")
def test_autostart_linux(qapp, qtbot):
    """Checks that autostart path is added correctly on Linux when setting is enabled."""
    setting = "Automatically start Vorta at login"

    # ensure file is present when autostart is enabled
    _click_toggle_setting(setting, qapp, qtbot)
    autostart_path = (
        Path(os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~") + '/.config') + "/autostart") / "vorta.desktop"
    )
    qtbot.waitUntil(lambda: autostart_path.exists(), **pytest._wait_defaults)
    with open(autostart_path) as desktop_file:
        desktop_file_text = desktop_file.read()
    assert desktop_file_text.startswith("[Desktop Entry]")

    # ensure file is removed when autostart is disabled
    _click_toggle_setting(setting, qapp, qtbot)
    if sys.platform == 'linux':
        assert not os.path.exists(autostart_path)


def test_enable_background_question(qapp, monkeypatch, mocker):
    """Tests that 'enable background question' correctly prompts user."""
    main = qapp.main_window
    close_event = Mock(value=QCloseEvent())

    # disable system trey and enable setting to test
    monkeypatch.setattr("vorta.views.main_window.is_system_tray_available", lambda: False)
    mocker.patch.object(vorta.store.models.SettingsModel, "get", return_value=Mock(value=True))
    mocker.patch.object(QMessageBox, "exec")  # prevent QMessageBox from stopping test

    # Create a mock for QMessageBox and its setText method
    mock_msgbox = mocker.Mock(spec=QMessageBox)
    mocker.patch("vorta.views.main_window.QMessageBox", return_value=mock_msgbox)

    main.closeEvent(close_event)

    mock_msgbox.setText.assert_called_once_with("Should Vorta continue to run in the background?")
    close_event.accept.assert_called_once()


def test_enable_fixed_units(qapp, qtbot, mocker):
    """Tests the 'enable fixed units' setting to ensure the archive tab sizes are displayed correctly."""
    tab = qapp.main_window.archiveTab
    setting = "Use the same unit of measurement for archive sizes"

    # set mocks
    mock_setting = mocker.patch.object(vorta.views.archive_tab.SettingsModel, "get", return_value=Mock(value=True))
    mock_pretty_bytes = mocker.patch.object(vorta.views.archive_tab, "pretty_bytes")

    # with setting enabled, fixed units should be determined and passed to pretty_bytes as an 'int'
    tab.populate_from_profile()
    mock_pretty_bytes.assert_called()
    kwargs_list = mock_pretty_bytes.call_args_list[0].kwargs
    assert 'fixed_unit' in kwargs_list
    assert isinstance(kwargs_list['fixed_unit'], int)

    # disable setting and reset mock
    mock_setting.return_value = Mock(value=False)
    mock_pretty_bytes.reset_mock()

    # with setting disabled, pretty_bytes should be called with fixed units set to 'None'
    tab.populate_from_profile()
    mock_pretty_bytes.assert_called()
    kwargs_list = mock_pretty_bytes.call_args_list[0].kwargs
    assert 'fixed_unit' in kwargs_list
    assert kwargs_list['fixed_unit'] is None

    # use the qt bot to click the setting and see that the refresh_archive emit works as intended.
    with qtbot.waitSignal(qapp.main_window.miscTab.refresh_archive, **pytest._wait_defaults):
        _click_toggle_setting(setting, qapp, qtbot)


@pytest.mark.skipif(sys.platform != 'darwin', reason="Full Disk Access check only on Darwin")
def test_check_full_disk_access(qapp, mocker):
    """Tests if the full disk access warning is properly silenced with the setting enabled"""

    # Set mocks for setting enabled
    mocker.patch.object(vorta.store.models.SettingsModel, "get", return_value=Mock(value=True))
    mocker.patch('pathlib.Path.exists', return_value=True)
    mocker.patch('os.access', return_value=False)
    mock_qmessagebox = mocker.patch('vorta.application.QMessageBox')

    # See that pop-up occurs
    qapp.check_darwin_permissions()
    mock_qmessagebox.assert_called()

    # Reset mocks for setting disabled
    mock_qmessagebox.reset_mock()
    mocker.patch.object(vorta.store.models.SettingsModel, "get", return_value=Mock(value=False))

    # See that pop-up does not occur
    qapp.check_darwin_permissions()
    mock_qmessagebox.assert_not_called()


def _click_toggle_setting(setting, qapp, qtbot):
    """Toggle setting checkbox in the misc tab"""
    miscTab = qapp.main_window.miscTab

    for x in range(miscTab.checkboxLayout.count()):
        item = miscTab.checkboxLayout.itemAt(x, QFormLayout.ItemRole.FieldRole)
        if item is not None:
            checkbox = item.itemAt(0).widget()
            if checkbox.text() == setting and isinstance(checkbox, QCheckBox):
                # Have to use pos to click checkbox correctly
                # https://stackoverflow.com/questions/19418125/pysides-qtest-not-checking-box/24070484#24070484
                pos = QtCore.QPoint(2, int(checkbox.height() / 2))
                qtbot.mouseClick(checkbox, QtCore.Qt.MouseButton.LeftButton, pos=pos)
                break



import pytest
from PyQt6.QtWidgets import QFormLayout, QWidget

from vorta.views.partials.password_input import PasswordInput, PasswordLineEdit


def test_create_password_line_edit(qtbot):
    password_line_edit = PasswordLineEdit()
    qtbot.addWidget(password_line_edit)
    assert password_line_edit is not None


def test_password_line_get_password(qtbot):
    password_line_edit = PasswordLineEdit()
    qtbot.addWidget(password_line_edit)

    assert password_line_edit.get_password() == ""

    qtbot.keyClicks(password_line_edit, "test")
    assert password_line_edit.get_password() == "test"


def test_password_line_visible(qtbot):
    password_line_edit = PasswordLineEdit()
    qtbot.addWidget(password_line_edit)
    assert not password_line_edit.visible

    password_line_edit.toggle_visibility()
    assert password_line_edit.visible

    with pytest.raises(TypeError):
        password_line_edit.visible = "OK"


def test_password_line_error_state(qtbot):
    password_line_edit = PasswordLineEdit()
    qtbot.addWidget(password_line_edit)
    assert not password_line_edit.error_state
    assert password_line_edit.styleSheet() == ""

    password_line_edit.error_state = True
    assert password_line_edit.error_state
    assert password_line_edit.styleSheet() == "QLineEdit { border: 2px solid red; }"


def test_password_line_visibility_button(qtbot):
    password_line_edit = PasswordLineEdit(show_visibility_button=False)
    qtbot.addWidget(password_line_edit)
    assert not password_line_edit._show_visibility_button

    password_line_edit = PasswordLineEdit()
    qtbot.addWidget(password_line_edit)
    assert password_line_edit._show_visibility_button

    # test visibility button
    password_line_edit.showHideAction.trigger()
    assert password_line_edit.visible
    password_line_edit.showHideAction.trigger()
    assert not password_line_edit.visible


# PasswordInput
def test_create_password_input(qapp, qtbot):
    password_input = PasswordInput()
    qtbot.addWidget(password_input.create_form_widget(parent=qapp.main_window))
    assert password_input is not None

    assert not password_input.passwordLineEdit.error_state
    assert not password_input.confirmLineEdit.error_state


def test_password_input_get_password(qapp, qtbot):
    password_input = PasswordInput()
    qtbot.addWidget(password_input.create_form_widget(parent=qapp.main_window))

    assert password_input.get_password() == ""

    password_input.passwordLineEdit.setText("test")
    assert password_input.get_password() == "test"


def test_password_input_validation(qapp, qtbot):
    password_input = PasswordInput(minimum_length=10)
    qtbot.addWidget(password_input.create_form_widget(parent=qapp.main_window))

    qtbot.keyClicks(password_input.passwordLineEdit, "123456789")
    qtbot.keyClicks(password_input.confirmLineEdit, "123456789")

    assert password_input.passwordLineEdit.error_state
    assert password_input.validation_label.text() == "Passwords must be at least 10 characters long."

    password_input.clear()
    qtbot.keyClicks(password_input.passwordLineEdit, "123456789")
    qtbot.keyClicks(password_input.confirmLineEdit, "test")

    assert password_input.passwordLineEdit.error_state
    assert password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == "Passwords must be identical and at least 10 characters long."

    password_input.clear()
    qtbot.keyClicks(password_input.passwordLineEdit, "1234567890")
    qtbot.keyClicks(password_input.confirmLineEdit, "test")

    assert not password_input.passwordLineEdit.error_state
    assert password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == "Passwords must be identical."

    password_input.clear()
    qtbot.keyClicks(password_input.passwordLineEdit, "1234567890")
    qtbot.keyClicks(password_input.confirmLineEdit, "1234567890")

    assert not password_input.passwordLineEdit.error_state
    assert not password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == ""


def test_password_input_validation_disabled(qapp, qtbot):
    password_input = PasswordInput(show_error=False)
    qtbot.addWidget(password_input.create_form_widget(parent=qapp.main_window))

    qtbot.keyClicks(password_input.passwordLineEdit, "test")
    qtbot.keyClicks(password_input.confirmLineEdit, "test")

    assert not password_input.passwordLineEdit.error_state
    assert not password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == ""

    password_input.set_validation_enabled(True)
    qtbot.keyClicks(password_input.passwordLineEdit, "s")
    qtbot.keyClicks(password_input.confirmLineEdit, "a")

    assert password_input.passwordLineEdit.error_state
    assert password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == "Passwords must be identical and at least 9 characters long."

    password_input.set_validation_enabled(False)
    assert not password_input.passwordLineEdit.error_state
    assert not password_input.confirmLineEdit.error_state
    assert password_input.validation_label.text() == ""


def test_password_input_set_label(qapp, qtbot):
    password_input = PasswordInput(label=["test", "test2"])
    qtbot.addWidget(password_input.create_form_widget(parent=qapp.main_window))

    assert password_input._label_password.text() == "test"
    assert password_input._label_confirm.text() == "test2"

    password_input.set_labels("test3", "test4")
    assert password_input._label_password.text() == "test3"
    assert password_input._label_confirm.text() == "test4"


def test_password_input_add_form_to_layout(qapp, qtbot):
    password_input = PasswordInput()

    widget = QWidget()
    form_layout = QFormLayout(widget)

    qtbot.addWidget(widget)
    password_input.add_form_to_layout(form_layout)

    assert form_layout.itemAt(0, QFormLayout.ItemRole.LabelRole).widget() == password_input._label_password
    assert form_layout.itemAt(0, QFormLayout.ItemRole.FieldRole).widget() == password_input.passwordLineEdit
    assert form_layout.itemAt(1, QFormLayout.ItemRole.LabelRole).widget() == password_input._label_confirm
    assert form_layout.itemAt(1, QFormLayout.ItemRole.FieldRole).widget() == password_input.confirmLineEdit



from collections import namedtuple

import psutil
import pytest
from PyQt6 import QtCore
from PyQt6.QtWidgets import QMenu
from test_constants import TEST_TEMP_DIR

import vorta.borg
import vorta.utils
import vorta.views.archive_tab
from vorta.store.models import ArchiveModel, BackupProfileModel


class MockFileDialog:
    def open(self, func):
        func()

    def selectedFiles(self):
        return [TEST_TEMP_DIR]


def test_prune_intervals(qapp, qtbot):
    prune_intervals = ['hour', 'day', 'week', 'month', 'year']
    main = qapp.main_window
    tab = main.archiveTab
    profile = BackupProfileModel.get(id=1)

    for i in prune_intervals:
        getattr(tab, f'prune_{i}').setValue(9)
        tab.save_prune_setting(None)
        profile = profile.refresh()
        assert getattr(profile, f'prune_{i}') == 9


def test_repo_list(qapp, qtbot, mocker, borg_json_output, archive_env):
    main, tab = archive_env

    stdout, stderr = borg_json_output('list')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    tab.refresh_archive_list()
    qtbot.waitUntil(lambda: not tab.bCheck.isEnabled(), **pytest._wait_defaults)
    assert not tab.bCheck.isEnabled()

    qtbot.waitUntil(lambda: 'Refreshing archives done.' in main.progressText.text(), **pytest._wait_defaults)
    assert ArchiveModel.select().count() == 6
    assert 'Refreshing archives done.' in main.progressText.text()
    assert tab.bCheck.isEnabled()


def test_repo_prune(qapp, qtbot, mocker, borg_json_output, archive_env):
    main, tab = archive_env

    stdout, stderr = borg_json_output('prune')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    qtbot.mouseClick(tab.bPrune, QtCore.Qt.MouseButton.LeftButton)

    qtbot.waitUntil(lambda: 'Refreshing archives done.' in main.progressText.text(), **pytest._wait_defaults)


def test_repo_compact(qapp, qtbot, mocker, borg_json_output, archive_env):
    vorta.utils.borg_compat.version = '1.2.0'
    main, tab = archive_env

    stdout, stderr = borg_json_output('compact')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    qtbot.mouseClick(tab.compactButton, QtCore.Qt.MouseButton.LeftButton)

    qtbot.waitUntil(
        lambda: 'compaction freed about 56.00 kB repository space' in main.logText.text(), **pytest._wait_defaults
    )
    vorta.utils.borg_compat.version = '1.1.0'


def test_check(qapp, mocker, borg_json_output, qtbot, archive_env):
    main, tab = archive_env

    stdout, stderr = borg_json_output('check')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    qtbot.mouseClick(tab.bCheck, QtCore.Qt.MouseButton.LeftButton)
    success_text = 'INFO: Archive consistency check complete'
    qtbot.waitUntil(lambda: success_text in main.logText.text(), **pytest._wait_defaults)


def test_mount(qapp, qtbot, mocker, borg_json_output, monkeypatch, choose_file_dialog, archive_env):
    def psutil_disk_partitions(**kwargs):
        DiskPartitions = namedtuple('DiskPartitions', ['device', 'mountpoint'])
        return [DiskPartitions('borgfs', TEST_TEMP_DIR)]

    monkeypatch.setattr(psutil, "disk_partitions", psutil_disk_partitions)
    main, tab = archive_env
    tab.archiveTable.selectRow(0)

    stdout, stderr = borg_json_output('prune')  # TODO: fully mock mount command?
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    monkeypatch.setattr(vorta.views.archive_tab, "choose_file_dialog", choose_file_dialog)

    tab.bmountarchive_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Mounted'), **pytest._wait_defaults)

    tab.bmountarchive_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Un-mounted successfully.'), **pytest._wait_defaults)

    tab.bmountrepo_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Mounted'), **pytest._wait_defaults)

    tab.bmountrepo_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Un-mounted successfully.'), **pytest._wait_defaults)


def test_archive_extract(qapp, qtbot, mocker, borg_json_output, archive_env):
    main, tab = archive_env
    tab.archiveTable.selectRow(0)
    stdout, stderr = borg_json_output('list_archive')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)
    tab.extract_action()

    qtbot.waitUntil(lambda: hasattr(tab, '_window'), **pytest._wait_defaults)

    model = tab._window.model
    assert model.root.children[0].subpath == 'home'
    assert 'test-archive, 2000' in tab._window.archiveNameLabel.text()


def test_archive_delete(qapp, qtbot, mocker, borg_json_output, archive_env):
    main, tab = archive_env

    tab.archiveTable.selectRow(0)
    stdout, stderr = borg_json_output('delete')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)
    mocker.patch.object(vorta.views.archive_tab.ArchiveTab, 'confirm_dialog', lambda x, y, z: True)
    tab.delete_action()
    qtbot.waitUntil(lambda: 'Archive deleted.' in main.progressText.text(), **pytest._wait_defaults)
    assert ArchiveModel.select().count() == 1
    assert tab.archiveTable.rowCount() == 1


def test_archive_copy(qapp, qtbot, monkeypatch, mocker, archive_env):
    main, tab = archive_env

    # mock the clipboard to ensure no changes are made to it during testing
    mocker.patch.object(qapp.clipboard(), "setMimeData")
    clipboard_spy = mocker.spy(qapp.clipboard(), "setMimeData")

    # test 'archive_copy()' by passing it an index to copy
    index = tab.archiveTable.model().index(0, 0)
    tab.archive_copy(index)
    assert clipboard_spy.call_count == 1
    actual_data = clipboard_spy.call_args[0][0]  # retrieves the QMimeData() object used in method call
    assert actual_data.text() == "test-archive"

    # test 'archive_copy()' by selecting a row to copy
    tab.archiveTable.selectRow(1)
    tab.archive_copy()
    assert clipboard_spy.call_count == 2
    actual_data = clipboard_spy.call_args[0][0]  # retrieves the QMimeData() object used in method call
    assert actual_data.text() == "test-archive1"


def test_refresh_archive_info(qapp, qtbot, mocker, borg_json_output, archive_env):
    main, tab = archive_env
    tab.archiveTable.selectRow(0)
    stdout, stderr = borg_json_output('info')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    with qtbot.waitSignal(tab.bRefreshArchive.clicked, timeout=5000):
        qtbot.mouseClick(tab.bRefreshArchive, QtCore.Qt.MouseButton.LeftButton)

    qtbot.waitUntil(lambda: tab.mountErrors.text() == 'Refreshed archives.', **pytest._wait_defaults)


def test_inline_archive_rename(qapp, qtbot, mocker, borg_json_output, archive_env):
    """
    Tests the functionality of in-line renaming an archive.
    """
    main, tab = archive_env

    tab.archiveTable.selectRow(0)
    new_archive_name = 'idf89d8f9d8fd98'
    stdout, stderr = borg_json_output('rename')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    # Trigger inline editing programmatically (more reliable than double-click simulation)
    item = tab.archiveTable.item(0, 4)
    tab.archiveTable.editItem(item)

    # Wait for edit mode to activate
    qtbot.waitUntil(lambda: tab.archiveTable.viewport().focusWidget() is not None, **pytest._wait_defaults)

    editor = tab.archiveTable.viewport().focusWidget()
    editor.setText(new_archive_name)
    qtbot.keyClick(editor, QtCore.Qt.Key.Key_Return)

    # Successful rename case
    qtbot.waitUntil(lambda: tab.archiveTable.model().index(0, 4).data() == new_archive_name, **pytest._wait_defaults)
    assert tab.archiveTable.model().index(0, 4).data() == new_archive_name


def test_archiveitem_contextmenu(qapp, qtbot, archive_env):
    main, tab = archive_env

    pos = tab.archiveTable.visualRect(tab.archiveTable.model().index(0, 0)).center()
    tab.archiveTable.customContextMenuRequested.emit(pos)
    qtbot.waitUntil(lambda: tab.archiveTable.findChild(QMenu) is not None, **pytest._wait_defaults)

    context_menu = tab.archiveTable.findChild(QMenu)

    assert context_menu is not None
    expected_actions = ['Copy', 'Recalculate', 'Mount…', 'Extract…', 'Rename…', 'Delete', 'Diff']
    for action in expected_actions:
        assert any(menu_actions.text() == action for menu_actions in context_menu.actions())



import pytest
from PyQt6 import QtCore


@pytest.mark.skip(reason="fails on macos with timeout when checking chromium-cache")
def test_exclusion_preview_populated(qapp, qtbot):
    main = qapp.main_window
    tab = main.sourceTab
    main.tabWidget.setCurrentIndex(1)

    qtbot.mouseClick(tab.bExclude, QtCore.Qt.MouseButton.LeftButton)
    qtbot.mouseClick(tab._window.bAddPattern, QtCore.Qt.MouseButton.LeftButton)

    qtbot.keyClicks(tab._window.customExclusionsList.viewport().focusWidget(), "custom pattern")
    qtbot.keyClick(tab._window.customExclusionsList.viewport().focusWidget(), QtCore.Qt.Key.Key_Enter)
    qtbot.waitUntil(lambda: "custom pattern" in tab._window.exclusionsPreviewText.toPlainText())

    tab._window.tabWidget.setCurrentIndex(1)

    tab._window.exclusionPresetsModel.itemFromIndex(tab._window.exclusionPresetsModel.index(0, 0)).setCheckState(
        QtCore.Qt.CheckState.Checked
    )

    qtbot.waitUntil(lambda: "# chromium-cache" in tab._window.exclusionsPreviewText.toPlainText())
    tab._window.tabWidget.setCurrentIndex(2)

    qtbot.keyClicks(tab._window.rawExclusionsText, "test raw pattern 1")
    qtbot.waitUntil(lambda: "test raw pattern 1\n" in tab._window.exclusionsPreviewText.toPlainText())

    qtbot.mouseClick(tab.bExclude, QtCore.Qt.MouseButton.LeftButton)
    qtbot.mouseClick(tab._window.bAddPatternExcludeIfPresent, QtCore.Qt.MouseButton.LeftButton)

    qtbot.keyClicks(tab._window.excludeIfPresentList.viewport().focusWidget(), "exclude_if_present_file")
    qtbot.keyClick(tab._window.excludeIfPresentList.viewport().focusWidget(), QtCore.Qt.Key.Key_Enter)
    qtbot.waitUntil(lambda: "exclude_if_present_file" in tab._window.exclusionsPreviewText.toPlainText())



import os
import time
from datetime import datetime as dt

import pytest
from peewee import SqliteDatabase
from test_constants import TEST_SOURCE_DIR, TEST_TEMP_DIR

import vorta
import vorta.application
import vorta.borg.jobs_manager
from vorta.store.models import (
    ArchiveModel,
    BackupProfileModel,
    EventLogModel,
    RepoModel,
    RepoPassword,
    SchemaVersion,
    SettingsModel,
    SourceFileModel,
    WifiSettingModel,
)
from vorta.views.main_window import ArchiveTab, MainWindow


def disconnect_all(signal):
    """
    Disconnect ALL handlers from a Qt signal.
    Unlike signal.disconnect() without arguments which only disconnects ONE handler,
    this function disconnects all connected handlers by calling disconnect in a loop
    until TypeError is raised (indicating no more handlers are connected).
    """
    while True:
        try:
            signal.disconnect()
        except TypeError:
            # No more handlers connected
            break


def all_workers_finished(jobs_manager):
    """
    Check if all worker threads have actually exited.
    This is more thorough than is_worker_running() which only checks current_job,
    because threads may still be alive briefly after current_job is set to None.
    """
    for worker in jobs_manager.workers.values():
        if worker.is_alive():
            return False
    return True


models = [
    RepoModel,
    RepoPassword,
    BackupProfileModel,
    SourceFileModel,
    SettingsModel,
    ArchiveModel,
    WifiSettingModel,
    EventLogModel,
    SchemaVersion,
]


def load_window(qapp: vorta.application.VortaApp):
    """
    Reload the main window of the given application.
    Used to repopulate fields after loading mock data.
    """
    qapp.main_window.deleteLater()
    # Skip QCoreApplication.processEvents() - it can trigger D-Bus operations that hang in CI.
    # Use a small sleep instead to allow deleteLater to be processed.
    time.sleep(0.1)
    del qapp.main_window
    qapp.main_window = MainWindow(qapp)


@pytest.fixture
def window_load(qapp):
    """
    A function to call to load fixture data into the app window.

    This is normally done by init_db, but if this fixture is used,
    the window load will be skipped to allow the test to load
    further data before then calling the returned function
    """
    return lambda: load_window(qapp)


@pytest.fixture(scope='function', autouse=True)
def init_db(qapp, qtbot, tmpdir_factory, request):
    tmp_db = tmpdir_factory.mktemp('Vorta').join('settings.sqlite')
    mock_db = SqliteDatabase(
        str(tmp_db),
        pragmas={
            'journal_mode': 'wal',
        },
    )
    vorta.store.connection.init_db(mock_db)

    # Force use of DB keyring instead of system keyring to avoid keychain prompts during tests
    keyring_setting = SettingsModel.get(key='use_system_keyring')
    keyring_setting.value = False
    keyring_setting.save()

    default_profile = BackupProfileModel(name='Default')
    default_profile.save()

    new_repo = RepoModel(url='i0fi93@i593.repo.borgbase.com:repo')
    new_repo.encryption = 'none'
    new_repo.save()

    default_profile.repo = new_repo.id
    default_profile.dont_run_on_metered_networks = False
    default_profile.validation_on = False
    default_profile.save()

    test_archive = ArchiveModel(snapshot_id='99999', name='test-archive', time=dt(2000, 1, 1, 0, 0), repo=1)
    test_archive.save()

    test_archive1 = ArchiveModel(snapshot_id='99998', name='test-archive1', time=dt(2000, 1, 1, 0, 0), repo=1)
    test_archive1.save()

    source_dir = SourceFileModel(dir=TEST_SOURCE_DIR, repo=new_repo, dir_size=100, dir_files_count=18, path_isdir=True)
    source_dir.save()

    # Disconnect ALL signal handlers before destroying main_window to avoid "deleted object" errors
    # Using disconnect_all() instead of disconnect() to ensure ALL handlers are removed,
    # not just one (which can leave stale connections from previous tests)
    disconnect_all(qapp.scheduler.schedule_changed)

    # Reload the window to apply the mock data
    # If this test has the `window_load` fixture,
    # it is responsible for calling this instead
    if 'window_load' not in request.fixturenames:
        load_window(qapp)

    yield

    # Teardown: cancel jobs and disconnect ALL signal handlers to prevent state leakage
    qapp.jobs_manager.cancel_all_jobs()

    # Wait for all worker threads to actually exit (not just for current_job to be None).
    # Use simple polling instead of qtbot.waitUntil to avoid Qt event loop hangs in CI.
    # qtbot.waitUntil processes Qt events while waiting, which can trigger D-Bus operations.
    timeout = pytest._wait_defaults.get('timeout', 20000) / 1000  # Convert ms to seconds
    start = time.time()
    while not all_workers_finished(qapp.jobs_manager):
        if time.time() - start > timeout:
            break
        time.sleep(0.1)

    # Skip QCoreApplication.processEvents() - it can trigger D-Bus operations that hang in CI

    # Disconnect signals
    disconnect_all(qapp.backup_finished_event)
    disconnect_all(qapp.scheduler.schedule_changed)

    # Clear the workers dict to prevent accumulation of dead thread references
    qapp.jobs_manager.workers.clear()
    qapp.jobs_manager.jobs.clear()
    mock_db.close()


@pytest.fixture
def choose_file_dialog(*args):
    class MockFileDialog:
        def __init__(self, *args, **kwargs):
            pass

        def open(self, func):
            func()

        def selectedFiles(self):
            return [TEST_TEMP_DIR]

    return MockFileDialog


@pytest.fixture
def borg_json_output():
    """
    Returns a function to read borg JSON output files.
    Opens real file handles (required for os.set_blocking and select.select),
    but tracks them for cleanup when the fixture is torn down.
    """
    open_files = []

    def _read_json(subcommand):
        stdout = open(f'tests/unit/borg_json_output/{subcommand}_stdout.json')
        stderr = open(f'tests/unit/borg_json_output/{subcommand}_stderr.json')
        open_files.append(stdout)
        open_files.append(stderr)
        return stdout, stderr

    yield _read_json

    # Clean up all opened files when the fixture is torn down
    for f in open_files:
        try:
            f.close()
        except Exception:
            pass


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture()
def archive_env(qapp, qtbot):
    """
    Common setup for unit tests involving the archive tab.
    """
    main: MainWindow = qapp.main_window
    tab: ArchiveTab = main.archiveTab
    main.tabWidget.setCurrentIndex(3)
    tab.populate_from_profile()
    qtbot.waitUntil(lambda: tab.archiveTable.rowCount() == 2, **pytest._wait_defaults)
    return main, tab



import os
import uuid
from typing import Any, Dict
from unittest.mock import MagicMock

import pytest
from PyQt6 import QtCore
from PyQt6.QtWidgets import QMessageBox

import vorta.borg.borg_job
from vorta.keyring.abc import VortaKeyring
from vorta.store.models import ArchiveModel, EventLogModel, RepoModel

LONG_PASSWORD = 'long-password-long'
SHORT_PASSWORD = 'hunter2'


@pytest.mark.parametrize(
    "first_password, second_password, validation_error",
    [
        (SHORT_PASSWORD, SHORT_PASSWORD, 'Passwords must be at least 9 characters long.'),
        (LONG_PASSWORD, SHORT_PASSWORD, 'Passwords must be identical.'),
        (SHORT_PASSWORD + "1", SHORT_PASSWORD, 'Passwords must be identical and at least 9 characters long.'),
        (LONG_PASSWORD, LONG_PASSWORD, ''),  # no error, password meets requirements.
    ],
)
def test_new_repo_password_validation(qapp, qtbot, borg_json_output, first_password, second_password, validation_error):
    # Add new repo window
    main = qapp.main_window
    tab = main.repoTab
    tab.new_repo()
    add_repo_window = tab._window
    qtbot.addWidget(add_repo_window)

    qtbot.keyClicks(add_repo_window.passwordInput.passwordLineEdit, first_password)
    qtbot.keyClicks(add_repo_window.passwordInput.confirmLineEdit, second_password)
    qtbot.mouseClick(add_repo_window.saveButton, QtCore.Qt.MouseButton.LeftButton)
    assert add_repo_window.passwordInput.validation_label.text() == validation_error


@pytest.mark.parametrize(
    "repo_name, error_text",
    [
        ('test_repo_name', ''),  # valid repo name
        ('a' * 64, ''),  # also valid (<=64 characters)
        ('a' * 65, 'Repository name must be less than 65 characters.'),  # not valid (>64 characters)
    ],
)
def test_repo_add_name_validation(qapp, qtbot, borg_json_output, repo_name, error_text):
    main = qapp.main_window
    tab = main.repoTab
    tab.new_repo()
    add_repo_window = tab._window
    test_repo_url = f'vorta-test-repo.{uuid.uuid4()}.com:repo'  # Random repo URL to avoid macOS keychain
    qtbot.addWidget(add_repo_window)

    qtbot.keyClicks(add_repo_window.repoURL, test_repo_url)
    qtbot.keyClicks(add_repo_window.repoName, repo_name)
    qtbot.mouseClick(add_repo_window.saveButton, QtCore.Qt.MouseButton.LeftButton)
    assert add_repo_window.errorText.text() == error_text


def test_repo_unlink(qapp, qtbot, monkeypatch):
    main = qapp.main_window
    tab = main.repoTab
    monkeypatch.setattr(QMessageBox, "show", lambda *args: True)

    # Assuming unlink action is the 1st in submenu
    tab.menuRepoUtil.actions()[0].trigger()

    qtbot.waitUntil(lambda: tab.repoSelector.count() == 1, **pytest._wait_defaults)
    assert RepoModel.select().count() == 0

    # Directly call create_backup_action and wait for signal
    with qtbot.waitSignal(qapp.backup_progress_event, timeout=5000):
        qapp.create_backup_action()

    assert 'Select a backup repository first.' in main.progressText.text()


def test_password_autofill(qapp, qtbot):
    main = qapp.main_window
    tab = main.repoTab
    tab.new_repo()
    add_repo_window = tab._window
    test_repo_url = f'vorta-test-repo.{uuid.uuid4()}.com:repo'  # Random repo URL to avoid macOS keychain

    keyring = VortaKeyring.get_keyring()
    password = str(uuid.uuid4())
    keyring.set_password('vorta-repo', test_repo_url, password)

    qtbot.keyClicks(add_repo_window.repoURL, test_repo_url)

    assert add_repo_window.passwordInput.passwordLineEdit.text() == password


def test_repo_add_failure(qapp, qtbot, borg_json_output):
    main = qapp.main_window
    tab = main.repoTab
    tab.new_repo()
    add_repo_window = tab._window
    qtbot.addWidget(add_repo_window)

    # Add repo with invalid URL
    qtbot.keyClicks(add_repo_window.repoURL, 'aaa')
    qtbot.mouseClick(add_repo_window.saveButton, QtCore.Qt.MouseButton.LeftButton)
    assert add_repo_window.errorText.text().startswith('Please enter a valid repo URL')


def test_repo_add_success(qapp, qtbot, mocker, borg_json_output):
    main = qapp.main_window
    tab = main.repoTab
    tab.new_repo()
    add_repo_window = tab._window
    test_repo_url = f'vorta-test-repo.{uuid.uuid4()}.com:repo'  # Random repo URL to avoid macOS keychain
    test_repo_name = 'Test Repo'

    # Enter valid repo URL, name, and password
    qtbot.keyClicks(add_repo_window.repoURL, test_repo_url)
    qtbot.keyClicks(add_repo_window.repoName, test_repo_name)
    qtbot.keyClicks(add_repo_window.passwordInput.passwordLineEdit, LONG_PASSWORD)
    qtbot.keyClicks(add_repo_window.passwordInput.confirmLineEdit, LONG_PASSWORD)

    stdout, stderr = borg_json_output('info')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    add_repo_window.run()
    qtbot.waitUntil(
        lambda: EventLogModel.select().where(EventLogModel.returncode == 0).count() == 2, **pytest._wait_defaults
    )

    assert RepoModel.get(id=2).url == test_repo_url

    keyring = VortaKeyring.get_keyring()
    assert keyring.get_password("vorta-repo", RepoModel.get(id=2).url) == LONG_PASSWORD
    assert tab.repoSelector.currentText() == f"{test_repo_name} - {test_repo_url}"


def test_ssh_dialog_success(qapp, qtbot, mocker, tmpdir):
    main = qapp.main_window
    tab = main.repoTab

    qtbot.mouseClick(tab.bAddSSHKey, QtCore.Qt.MouseButton.LeftButton)
    ssh_dialog = tab._window
    ssh_dialog_closed = mocker.spy(ssh_dialog, 'reject')
    ssh_dir = tmpdir
    key_tmpfile = ssh_dir.join("id_rsa-test")
    pub_tmpfile = ssh_dir.join("id_rsa-test.pub")
    key_tmpfile_full = os.path.join(key_tmpfile.dirname, key_tmpfile.basename)
    ssh_dialog.outputFileTextBox.setText(key_tmpfile_full)
    ssh_dialog.generate_key()

    # Ensure new key file was created
    qtbot.waitUntil(lambda: ssh_dialog_closed.called, **pytest._wait_defaults)
    assert len(ssh_dir.listdir()) == 2

    # Ensure new key is populated in SSH combobox
    mocker.patch('os.path.expanduser', return_value=str(tmpdir))
    tab.init_ssh()
    assert tab.sshComboBox.count() == 2

    # Ensure valid keys were created
    key_tmpfile_content = key_tmpfile.read()
    assert key_tmpfile_content.startswith('-----BEGIN OPENSSH PRIVATE KEY-----')
    pub_tmpfile_content = pub_tmpfile.read()
    assert pub_tmpfile_content.startswith('ssh-ed25519')


def test_ssh_dialog_failure(qapp, qtbot, mocker, monkeypatch, tmpdir):
    main = qapp.main_window
    tab = main.repoTab
    monkeypatch.setattr(QMessageBox, "show", lambda *args: True)
    failure_message = mocker.spy(tab, "create_ssh_key_failure")

    qtbot.mouseClick(tab.bAddSSHKey, QtCore.Qt.MouseButton.LeftButton)
    ssh_dialog = tab._window
    ssh_dir = tmpdir
    key_tmpfile = ssh_dir.join("invalid///===for_testing")
    key_tmpfile_full = os.path.join(key_tmpfile.dirname, key_tmpfile.basename)
    ssh_dialog.outputFileTextBox.setText(key_tmpfile_full)
    ssh_dialog.generate_key()

    qtbot.waitUntil(lambda: failure_message.called, **pytest._wait_defaults)
    failure_message.assert_called_once()

    # Ensure no new ney file was created
    assert len(ssh_dir.listdir()) == 0

    # Ensure no new key file in combo box
    mocker.patch('os.path.expanduser', return_value=str(tmpdir))
    tab.init_ssh()
    assert tab.sshComboBox.count() == 1


def test_ssh_copy_to_clipboard_action(qapp, qtbot, mocker, tmpdir):
    """Testing the proper QMessageBox dialogue appears depending on the copy action circumstances."""
    tab = qapp.main_window.repoTab

    # set mocks to test assertions and prevent test interruptions
    text = mocker.patch.object(QMessageBox, "setText")
    mocker.patch.object(QMessageBox, "show")
    mocker.patch.object(qapp.clipboard(), "setText")

    qtbot.mouseClick(tab.bAddSSHKey, QtCore.Qt.MouseButton.LeftButton)
    ssh_dialog = tab._window
    ssh_dialog_closed = mocker.spy(ssh_dialog, 'reject')
    ssh_dir = tmpdir
    key_tmpfile = ssh_dir.join("id_rsa-test")
    pub_tmpfile = ssh_dir.join("id_rsa-test.pub")
    key_tmpfile_full = os.path.join(key_tmpfile.dirname, key_tmpfile.basename)
    ssh_dialog.outputFileTextBox.setText(key_tmpfile_full)
    ssh_dialog.generate_key()

    # Ensure new key file was created
    qtbot.waitUntil(lambda: ssh_dialog_closed.called, **pytest._wait_defaults)
    assert len(ssh_dir.listdir()) == 2
    # populate the ssh combobox with the ssh key we created in tmpdir
    mock_expanduser = mocker.patch('os.path.expanduser', return_value=str(tmpdir))
    tab.init_ssh()
    assert tab.sshComboBox.count() == 2

    # test when no ssh key is selected to copy
    assert tab.sshComboBox.currentIndex() == 0
    qtbot.mouseClick(tab.sshKeyToClipboardButton, QtCore.Qt.MouseButton.LeftButton)
    message = "Select a public key from the dropdown first."
    text.assert_called_with(message)

    # Select a key and copy it
    mock_expanduser.return_value = pub_tmpfile
    tab.sshComboBox.setCurrentIndex(1)
    assert tab.sshComboBox.currentIndex() == 1
    qtbot.mouseClick(tab.sshKeyToClipboardButton, QtCore.Qt.MouseButton.LeftButton)
    message = "The selected public SSH key was copied to the clipboard. Use it to set up remote repo permissions."
    text.assert_called_with(message)

    # handle ssh key file not found
    mock_expanduser.return_value = "foobar"
    assert tab.sshComboBox.currentIndex() == 1
    qtbot.mouseClick(tab.sshKeyToClipboardButton, QtCore.Qt.MouseButton.LeftButton)
    message = "Could not find public key."
    text.assert_called_with(message)


def test_create(qapp, borg_json_output, mocker, qtbot):
    main = qapp.main_window
    stdout, stderr = borg_json_output('create')
    popen_result = mocker.MagicMock(stdout=stdout, stderr=stderr, returncode=0)
    mocker.patch.object(vorta.borg.borg_job, 'Popen', return_value=popen_result)

    qtbot.mouseClick(main.createStartBtn, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: 'Backup finished.' in main.progressText.text(), **pytest._wait_defaults)
    qtbot.waitUntil(lambda: main.createStartBtn.isEnabled(), **pytest._wait_defaults)
    assert EventLogModel.select().count() == 1
    assert ArchiveModel.select().count() == 3
    assert RepoModel.get(id=1).unique_size == 15520474
    assert main.createStartBtn.isEnabled()
    assert main.archiveTab.archiveTable.rowCount() == 3
    assert main.scheduleTab.logPage.logPage.rowCount() == 1


@pytest.mark.parametrize(
    "response",
    [
        {
            "return_code": 0,  # no error
            "error": "",
            "icon": None,
            "info": None,
        },
        {
            "return_code": 1,  # warning
            "error": "Borg exited with warning status (rc 1).",
            "icon": QMessageBox.Icon.Warning,
            "info": "",
        },
        {
            "return_code": 2,  # critical error
            "error": "Repository data check for repo test_repo_url failed. Error code 2",
            "icon": QMessageBox.Icon.Critical,
            "info": "Consider repairing or recreating the repository soon to avoid missing data.",
        },
        {
            "return_code": 135,  # 128 + n = kill signal n
            "error": "killed by signal 7",
            "icon": QMessageBox.Icon.Critical,
            "info": "The process running the check job got a kill signal. Try again.",
        },
        {"return_code": 130, "error": "", "icon": None, "info": None},  # keyboard interrupt
    ],
)
def test_repo_check_failed_response(qapp, qtbot, mocker, response):
    """Test the processing of the signal that a repo consistency check has failed."""
    mock_result: Dict[str, Any] = {
        'params': {'repo_url': 'test_repo_url'},
        'returncode': response["return_code"],
        'errors': [(0, 'test_error_message')] if response["return_code"] not in [0, 130] else None,
    }

    mock_exec = mocker.patch.object(QMessageBox, "exec")
    mock_text = mocker.patch.object(QMessageBox, "setText")
    mock_info = mocker.patch.object(QMessageBox, "setInformativeText")
    mock_icon = mocker.patch.object(QMessageBox, "setIcon")

    qapp.check_failed_response(mock_result)

    # return codes 0 and 130 do not provide a message
    # for all other return codes, assert the message is formatted correctly
    if mock_exec.call_count != 0:
        mock_icon.assert_called_with(response["icon"])
        assert response["error"] in mock_text.call_args[0][0]
        assert response["info"] in mock_info.call_args[0][0]


def test_repo_change_passphrase_action(qapp, mocker):
    """Test that the ChangeBorgPassphraseWindow is opened and the signal is connected."""
    tab = qapp.main_window.repoTab

    mock_profile = MagicMock()
    mock_profile.repo.encryption = 'repokey-blake2'
    tab.profile = MagicMock(return_value=mock_profile)

    mock_window_cls = mocker.patch('vorta.views.repo_tab.ChangeBorgPassphraseWindow')
    mock_window = mock_window_cls.return_value

    tab.repo_change_passphrase_action()

    mock_window_cls.assert_called_once_with(mock_profile)
    mock_window.setParent.assert_called_once_with(tab, QtCore.Qt.WindowType.Sheet)
    mock_window.open.assert_called_once()
    mock_window.change_borg_passphrase.connect.assert_called_once_with(tab._handle_passphrase_change_result)


@pytest.mark.parametrize(
    "result, expected_title, expected_text",
    [
        (
            {"returncode": 0},
            "Passphrase Changed",
            "The borg passphrase was successfully changed.",
        ),
        (
            {"returncode": 1},
            "Passphrase Change Failed",
            "Unable to change the repository passphrase. Please try again.",
        ),
    ],
)
def test_handle_passphrase_change_result(qapp, qtbot, mocker, result, expected_title, expected_text):
    """Test the _handle_passphrase_change_result method for both success and failure cases."""
    main = qapp.main_window
    tab = main.repoTab

    mock_msgbox = mocker.patch('vorta.views.repo_tab.QMessageBox', autospec=True)
    mock_instance = mock_msgbox.return_value

    tab._handle_passphrase_change_result(result)

    mock_instance.setWindowTitle.assert_called_once_with(tab.tr(expected_title))
    mock_instance.setText.assert_called_once_with(tab.tr(expected_text))
    mock_instance.show.assert_called_once()



"""
These tests compare the output of the diff command with the expected output.
"""

import pytest
from packaging.version import Version

import vorta.borg
import vorta.utils
import vorta.views.archive_tab
from vorta.borg.diff import BorgDiffJob
from vorta.views.diff_result import (
    ChangeType,
    DiffTree,
    FileType,
    ParseThread,
)


@pytest.mark.parametrize(
    "archive_name_1, archive_name_2, expected",
    [
        (
            "test-archive1",
            "test-archive2",
            [
                {
                    "subpath": "dir",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                        "modified": None,
                    },
                    "min_version": "1.2.4",
                },
                {
                    "subpath": "file",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                        "modified": (0, 0),
                    },
                    "min_version": "1.2.4",
                    "max_version": "1.4.0",
                },
                {
                    "subpath": "file",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                        "modified": None,
                    },
                    "min_version": "1.4.1",
                },
                {
                    "subpath": "chrdev",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.ADDED,
                        "modified": None,
                    },
                },
                {
                    "subpath": "fifo",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.ADDED,
                        "modified": None,
                    },
                },
                {
                    "subpath": "hardlink",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.ADDED,
                        "modified": None,
                    },
                },
                {
                    "subpath": "symlink",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.ADDED,
                        "modified": None,
                    },
                },
            ],
        ),
        (
            "test-archive2",
            "test-archive3",
            [
                {
                    "subpath": "borg_src",
                    "match_startsWith": True,
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                        "modified": None,
                    },
                    "min_version": "1.2.4",
                },
                {
                    "subpath": "dir",
                    "data": {
                        "file_type": FileType.DIRECTORY,
                        "change_type": ChangeType.REMOVED,
                        "modified": None,
                    },
                },
                {
                    "subpath": "chrdev",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "fifo",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "file",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "hardlink",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "symlink",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "dir1",
                    "data": {
                        "file_type": FileType.DIRECTORY,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "chrdev",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "fifo",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "file",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "hardlink",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "symlink",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.ADDED,
                    },
                },
            ],
        ),
        (
            "test-archive3",
            "test-archive4",
            [
                {
                    "subpath": "dir1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                    },
                    "min_version": "1.2.4",
                },
                {
                    "subpath": "chrdev",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "chrdev1",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "fifo",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "fifo1",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "file",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "file1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "hardlink",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "hardlink1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.ADDED,
                    },
                },
                {
                    "subpath": "symlink",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "symlink1",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.ADDED,
                    },
                },
            ],
        ),
        (
            "test-archive4",
            "test-archive5",
            [
                {
                    "subpath": "dir1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                    },
                    "min_version": "1.2.4",
                },
                {
                    "subpath": "chrdev1",
                    "data": {
                        "file_type": FileType.CHRDEV,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "fifo1",
                    "data": {
                        "file_type": FileType.FIFO,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "file1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "hardlink1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.REMOVED,
                    },
                },
                {
                    "subpath": "symlink1",
                    "data": {
                        "file_type": FileType.LINK,
                        "change_type": ChangeType.REMOVED,
                    },
                },
            ],
        ),
        (
            "test-archive5",
            "test-archive6",
            [
                {
                    "subpath": "dir1",
                    "data": {
                        "file_type": FileType.FILE,
                        "change_type": ChangeType.MODIFIED,
                    },
                    "min_version": "1.2.4",
                },
            ],
        ),
    ],
)
def test_archive_diff_lines(
    qapp,
    qtbot,
    borg_version,
    create_test_repo,
    archive_name_1,
    archive_name_2,
    expected,
):
    """Test that the diff lines are parsed correctly for supported borg versions"""
    parsed_borg_version = borg_version[1]
    supports_fifo = parsed_borg_version > Version("1.1.18")
    supports_chrdev = create_test_repo[2]

    params = BorgDiffJob.prepare(
        vorta.store.models.BackupProfileModel.select().first(),
        archive_name_1,
        archive_name_2,
    )
    thread = BorgDiffJob(params["cmd"], params, qapp)

    with qtbot.waitSignal(thread.result, **pytest._wait_defaults) as blocker:
        blocker.connect(thread.updated)
        thread.run()

    diff_lines = blocker.args[0]["data"]
    json_lines = blocker.args[0]["params"]["json_lines"]

    model = DiffTree()
    model.setMode(model.DisplayMode.FLAT)

    # Use ParseThread to parse the diff lines
    parse_thread = ParseThread(diff_lines, json_lines, model)
    parse_thread.start()
    qtbot.waitUntil(lambda: parse_thread.isFinished(), **pytest._wait_defaults)

    expected = [
        item
        for item in expected
        if (
            ("min_version" not in item or Version(item["min_version"]) <= parsed_borg_version)
            and ("max_version" not in item or Version(item["max_version"]) >= parsed_borg_version)
            and (item["data"]["file_type"] != FileType.FIFO or supports_fifo)
            and (item["data"]["file_type"] != FileType.CHRDEV or supports_chrdev)
        )
    ]

    # diff versions of borg produce inconsistent ordering of diff lines so we sort the expected and model
    expected = sorted(expected, key=lambda item: item["subpath"])
    sorted_model = sorted(
        [model.index(index, 0).internalPointer() for index in range(model.rowCount())],
        key=lambda item: item.subpath,
    )

    assert len(sorted_model) == len(expected)

    for index, item in enumerate(expected):
        if "match_startsWith" in item and item["match_startsWith"]:
            assert sorted_model[index].subpath.startswith(item["subpath"])
        else:
            assert sorted_model[index].subpath == item["subpath"]

        for key, value in item["data"].items():
            assert getattr(sorted_model[index].data, key) == value



"""
Test initialization of new repositories and adding existing ones.
"""

import os
from pathlib import PurePath

import pytest
from PyQt6.QtCore import QCoreApplication, Qt
from PyQt6.QtWidgets import QMessageBox

import vorta.borg
import vorta.utils
import vorta.views.repo_add_dialog
from tests.integration.conftest import all_workers_finished

LONG_PASSWORD = 'long-password-long'
TEST_REPO_NAME = 'TEST - REPONAME'


def test_create_repo(qapp, qtbot, monkeypatch, choose_file_dialog, tmpdir):
    """Test initializing a new repository"""
    main = qapp.main_window
    main.repoTab.new_repo()
    add_repo_window = main.repoTab._window
    main.show()

    # create new folder in tmpdir
    new_repo_path = tmpdir.join('new_repo')
    new_repo_path.mkdir()

    monkeypatch.setattr(
        vorta.views.repo_add_dialog,
        "choose_file_dialog",
        lambda *args, **kwargs: choose_file_dialog(*args, **kwargs, subdirectory=new_repo_path.basename),
    )
    qtbot.mouseClick(add_repo_window.chooseLocalFolderButton, Qt.MouseButton.LeftButton)

    # clear auto input of repo name from url
    add_repo_window.repoName.selectAll()
    add_repo_window.repoName.del_()
    qtbot.keyClicks(add_repo_window.repoName, TEST_REPO_NAME)

    qtbot.keyClicks(add_repo_window.passwordInput.passwordLineEdit, LONG_PASSWORD)
    qtbot.keyClicks(add_repo_window.passwordInput.confirmLineEdit, LONG_PASSWORD)

    initial_count = main.repoTab.repoSelector.count()
    add_repo_window.run()

    # Wait for all worker threads to fully exit (more thorough than is_worker_running)
    qtbot.waitUntil(lambda: all_workers_finished(qapp.jobs_manager), **pytest._wait_defaults)

    # Process pending Qt events to ensure signals are delivered and UI is updated
    QCoreApplication.processEvents()

    # Wait for the new repo to appear in the selector
    qtbot.waitUntil(lambda: main.repoTab.repoSelector.count() == initial_count + 1, **pytest._wait_defaults)

    # Check if repo was created in tmpdir
    repo_url = (
        vorta.store.models.RepoModel.select().where(vorta.store.models.RepoModel.name == TEST_REPO_NAME).get().url
    )
    assert PurePath(repo_url).parent == tmpdir
    assert PurePath(repo_url).name == 'new_repo'

    # check that new_repo_path contains folder data
    assert os.path.exists(new_repo_path.join('data'))
    assert os.path.exists(new_repo_path.join('config'))
    assert os.path.exists(new_repo_path.join('README'))


def test_add_existing_repo(qapp, qtbot, monkeypatch, choose_file_dialog):
    """Test adding an existing repository"""
    main = qapp.main_window
    tab = main.repoTab

    main.tabWidget.setCurrentIndex(0)
    current_repo_path = vorta.store.models.RepoModel.select().first().url

    monkeypatch.setattr(QMessageBox, "show", lambda *args: True)

    # Assuming unlink action is the first in submenu
    tab.menuRepoUtil.actions()[0].trigger()

    qtbot.waitUntil(
        lambda: tab.repoSelector.count() == 1 and tab.repoSelector.currentText() == "No repository selected",
        **pytest._wait_defaults,
    )

    # add existing repo again
    main.repoTab.add_existing_repo()
    add_repo_window = main.repoTab._window

    monkeypatch.setattr(
        vorta.views.repo_add_dialog,
        "choose_file_dialog",
        lambda *args, **kwargs: choose_file_dialog(*args, **kwargs, directory=current_repo_path),
    )
    qtbot.mouseClick(add_repo_window.chooseLocalFolderButton, Qt.MouseButton.LeftButton)

    # clear auto input of repo name from url
    add_repo_window.repoName.selectAll()
    add_repo_window.repoName.del_()
    qtbot.keyClicks(add_repo_window.repoName, TEST_REPO_NAME)

    add_repo_window.run()

    # Wait for all worker threads to fully exit (more thorough than is_worker_running)
    qtbot.waitUntil(lambda: all_workers_finished(qapp.jobs_manager), **pytest._wait_defaults)

    # Process pending Qt events to ensure signals are delivered and UI is updated
    QCoreApplication.processEvents()

    # Wait for the repo to appear in the selector
    # After unlink, count is 1 (only placeholder). After adding repo, count should be 2.
    qtbot.waitUntil(lambda: tab.repoSelector.count() == 2, **pytest._wait_defaults)

    # check that repo was added correctly
    assert vorta.store.models.RepoModel.select().first().url == str(current_repo_path)
    assert vorta.store.models.RepoModel.select().first().name == TEST_REPO_NAME



import sys
import uuid

import pytest

from vorta.keyring.abc import VortaKeyring
from vorta.utils import (
    find_best_unit_for_sizes,
    get_path_datasize,
    is_system_tray_available,
    normalize_path,
    pretty_bytes,
    sort_sizes,
)


def test_keyring():
    UNICODE_PW = 'kjalsdfüadsfäadsfß'
    REPO = f'ssh://asdf123@vorta-test-repo.{uuid.uuid4()}.com/./repo'  # Random repo URL

    keyring = VortaKeyring.get_keyring()
    keyring.set_password('vorta-repo', REPO, UNICODE_PW)
    assert keyring.get_password("vorta-repo", REPO) == UNICODE_PW


@pytest.mark.parametrize(
    "input_sizes, expected_sorted",
    [
        # Basic ordering
        (["1.0 GB", "2.0 MB", "3.0 KB"], ["3.0 KB", "2.0 MB", "1.0 GB"]),
        # Multiple same units
        (["3.0 GB", "2.0 GB", "1.0 GB"], ["1.0 GB", "2.0 GB", "3.0 GB"]),
        # Multiple different units
        (["2.0 MB", "3.0 GB", "1.0 KB", "5.0 GB"], ["1.0 KB", "2.0 MB", "3.0 GB", "5.0 GB"]),
        # Larger to smaller units
        (["1.0 YB", "1.0 ZB", "1.0 EB", "1.0 PB"], ["1.0 PB", "1.0 EB", "1.0 ZB", "1.0 YB"]),
        # Skipping non-numeric sizes
        (["2x MB", "3.0 KB", "apple GB", "1.0 GB"], ["3.0 KB", "1.0 GB"]),
        # Skipping invalid suffix
        (["1.0 XX", "5.0 YY", "9.0 ZZ", "1.0 MB"], ["1.0 MB"]),
        # Floats with decimals
        (["2.5 GB", "2.3 GB", "1.1 MB"], ["1.1 MB", "2.3 GB", "2.5 GB"]),
        # Checking the same sizes across different units
        (["1.0 MB", "1000.0 KB"], ["1000.0 KB", "1.0 MB"]),
        # Handle empty lists
        ([], []),
    ],
)
def test_sort_sizes(input_sizes, expected_sorted):
    assert sort_sizes(input_sizes) == expected_sorted


@pytest.mark.parametrize(
    "precision, expected_unit",
    [
        (0, 1),  # return units as "1" (represents KB), min=100KB
        (1, 2),  # return units as "2" (represents MB), min=0.1MB
        (2, 2),  # still returns KB, since 0.1MB < min=0.001 GB to allow for GB to be best_unit
    ],
)
def test_best_unit_for_sizes_precision(precision, expected_unit):
    MB = 1000000
    sizes = [int(0.1 * MB), 100 * MB, 2000 * MB]
    best_unit = find_best_unit_for_sizes(sizes, metric=True, precision=precision)
    assert best_unit == expected_unit


@pytest.mark.parametrize(
    "sizes, expected_unit",
    [
        ([], 0),  # no sizes given but should still return "0" (represents bytes) as best representation
        ([102], 0),  # non-metric size 102 < 0.1KB (102 < 0.1 * 1024), so it will return 0 instead of 1
        ([103], 1),  # non-metric size 103 > 0.1KB (103 < 0.1 * 1024), so it will return 1
    ],
)
def test_best_unit_for_sizes_nonmetric(sizes, expected_unit):
    best_unit = find_best_unit_for_sizes(sizes, metric=False, precision=1)
    assert best_unit == expected_unit


@pytest.mark.parametrize(
    "size, metric, precision, fixed_unit, expected_output",
    [
        (10**5, True, 1, 2, "0.1 MB"),  # 100KB, metric, precision 1, fixed unit "2" (MB)
        (10**6, True, 0, 2, "1 MB"),  # 1MB, metric, precision 0, fixed unit "2" (MB)
        (10**6, True, 1, 2, "1.0 MB"),  # 1MB, metric, precision 1, fixed unit "2" (MB)
        (1024 * 1024, False, 1, 2, "1.0 MiB"),  # 1MiB, nonmetric, precision 1, fixed unit "2" (MiB)
    ],
)
def test_pretty_bytes_fixed_units(size, metric, precision, fixed_unit, expected_output):
    """
    Test pretty bytes when specifying a fixed unit of measurement
    """
    output = pretty_bytes(size, metric=metric, precision=precision, fixed_unit=fixed_unit)
    assert output == expected_output


@pytest.mark.parametrize(
    "size, metric, expected_output",
    [
        (10**6, True, "1.0 MB"),  # 1MB, metric
        (10**24, True, "1.0 YB"),  # 1YB, metric
        (10**30, True, "1000000.0 YB"),  # test huge number, metric
        (1024 * 1024, False, "1.0 MiB"),  # 1MiB, nonmetric
        (2**40 * 2**40, False, "1.0 YiB"),  # 1YiB, nonmetric
    ],
)
def test_pretty_bytes_nonfixed_units(size, metric, expected_output):
    # test pretty bytes when NOT specifying a fixed unit of measurement
    output = pretty_bytes(size, metric=metric, precision=1)
    assert output == expected_output


def test_normalize_path():
    """
    Test that path is normalized for macOS, but does nothing for other platforms.
    """
    input_path = '/Users/username/caf\u00e9/file.txt'
    expected_output = '/Users/username/café/file.txt'

    actual_output = normalize_path(input_path)

    if sys.platform == 'darwin':
        assert actual_output == expected_output
    else:
        assert actual_output == input_path


def test_get_path_datasize(tmpdir):
    """
    Test that get_path_datasize() works correctly when passed excluded patterns.
    """
    # Create a temporary directory for testing
    test_dir = tmpdir.mkdir("test_dir")
    test_file = test_dir.join("test_file.txt")
    test_file.write("Hello, World!")

    # Create a subdirectory with a file to exclude
    excluded_dir = test_dir.mkdir("excluded_dir")
    excluded_file = excluded_dir.join("excluded_file.txt")
    excluded_file.write("Excluded file, should not be checked.")

    exclude_patterns = [f"{excluded_dir}"]

    # Test when the path is a directory
    data_size, files_count = get_path_datasize(str(test_dir), exclude_patterns)
    assert data_size == len("Hello, World!")
    assert files_count == 1

    # Test when the path is a file
    data_size, files_count = get_path_datasize(str(test_file), exclude_patterns)
    assert data_size == len("Hello, World!")
    assert files_count == 1

    # Test when the path is a directory with an excluded file
    data_size, files_count = get_path_datasize(str(excluded_dir), exclude_patterns)
    assert data_size == 0
    assert files_count == 0


def test_is_system_tray_available(mocker):
    """
    Sanity check to ensure proper behavior
    """
    mocker.patch('PyQt6.QtWidgets.QSystemTrayIcon.isSystemTrayAvailable', return_value=False)
    assert is_system_tray_available() is False
    mocker.patch('PyQt6.QtWidgets.QSystemTrayIcon.isSystemTrayAvailable', return_value=True)
    assert is_system_tray_available() is True



from unittest.mock import MagicMock, patch

import pytest
from PyQt6.QtCore import QDir, QModelIndex
from PyQt6.QtWidgets import QMessageBox

from vorta.filedialog import VortaFileDialog, VortaFileSelector


@pytest.fixture
def file_dialog(qapp):
    dialog = VortaFileDialog()
    yield dialog
    dialog.close()


def test_dialog_initial_state(file_dialog):
    """Verify the dialog initializes with correct defaults."""
    assert file_dialog.windowTitle() == "Vorta File Dialog"
    assert file_dialog.path_bar.text() == QDir.homePath()
    assert file_dialog.tree.model() == file_dialog.model
    assert file_dialog.tree.rootIndex() == file_dialog.model.index(QDir.homePath())


def test_selected_paths_with_access(file_dialog, mocker):
    """Test path selection when read access is granted."""
    mock_index = mocker.MagicMock()
    mock_index.column.return_value = 0
    mock_selection = mocker.MagicMock()
    mock_selection.selectedIndexes.return_value = [mock_index]

    with (
        patch.object(file_dialog.tree, 'selectionModel', return_value=mock_selection),
        patch.object(file_dialog.model, 'filePath', return_value="/valid/path"),
        patch('os.access', return_value=True),
    ):
        result = file_dialog.selected_paths()
        assert result == ["/valid/path"]


def test_selected_paths_without_access(file_dialog, mocker):
    """Test path selection when read access is denied."""
    mock_index = mocker.MagicMock(spec=QModelIndex)
    mock_index.column.return_value = 0
    mock_selection = mocker.MagicMock()
    mock_selection.selectedIndexes.return_value = [mock_index]

    with (
        patch.object(file_dialog.tree, 'selectionModel', return_value=mock_selection),
        patch.object(file_dialog.model, 'filePath', return_value="/no/access"),
        patch('os.access', return_value=False),
        patch('vorta.filedialog.QMessageBox') as MockQMessageBox,
    ):
        # mocking entire QMessageBox as msg.exec() was causing race conditions
        mock_msg = MockQMessageBox.return_value
        mock_msg.exec.return_value = QMessageBox.StandardButton.Ok

        result = file_dialog.selected_paths()

        assert result == []
        MockQMessageBox.assert_called_once()
        mock_msg.exec.assert_called_once()


def test_path_changed_valid(file_dialog, qtbot):
    """Test path bar updates with a valid path."""
    test_path = QDir.homePath()

    with patch('os.path.exists', return_value=True):
        file_dialog.path_bar.setText(test_path)
        qtbot.waitUntil(lambda: file_dialog.path_bar.styleSheet() == "")

        assert file_dialog.tree.rootIndex() == file_dialog.model.index(test_path)


def test_path_changed_invalid(file_dialog, qtbot):
    """Test path bar updates with an invalid path."""
    with patch('os.path.exists', return_value=False):
        file_dialog.path_bar.setText("/invalid/path")
        qtbot.waitUntil(lambda: file_dialog.path_bar.styleSheet() == "background-color: #ffcccc")


def test_file_selector_get_paths(mocker):
    """Test the instance selector method with successful selection."""
    mock_dialog = mocker.patch('vorta.filedialog.VortaFileDialog')
    mock_instance = mock_dialog.return_value
    mock_instance.exec.return_value = True
    mock_instance.selected_paths.return_value = ["/test/path"]

    selector = VortaFileSelector()
    paths = selector.get_paths()
    assert paths == ["/test/path"]


def test_file_selector_cancel(mocker):
    """Test the instance selector method when cancelled."""
    mock_dialog = mocker.patch('vorta.filedialog.VortaFileDialog')
    mock_instance = mock_dialog.return_value
    mock_instance.exec.return_value = False

    selector = VortaFileSelector()
    paths = selector.get_paths()
    assert paths == []



"""
Shared constants for unit tests.
These constants provide cross-platform compatible paths instead of hardcoded /tmp paths.
"""

import os
import tempfile

# Use system temp directory for test paths instead of hardcoded /tmp for cross-platform compatibility
TEST_TEMP_DIR = tempfile.gettempdir()
TEST_SOURCE_DIR = os.path.join(TEST_TEMP_DIR, 'test_source')



"""
This file contains tests for the Archive tab to test the various archive related borg commands.
"""

import sys
from collections import namedtuple

import psutil
import pytest
from PyQt6 import QtCore

import vorta.borg
import vorta.utils
import vorta.views.archive_tab
from vorta.store.models import ArchiveModel


def test_repo_list(qapp, qtbot):
    """Test that the archives are created and repo list is populated correctly"""
    main = qapp.main_window
    tab = main.archiveTab
    main.tabWidget.setCurrentIndex(3)
    tab.refresh_archive_list()
    qtbot.waitUntil(lambda: not tab.bCheck.isEnabled(), **pytest._wait_defaults)
    assert not tab.bCheck.isEnabled()

    qtbot.waitUntil(lambda: 'Refreshing archives done.' in main.progressText.text(), **pytest._wait_defaults)
    assert ArchiveModel.select().count() == 6
    assert 'Refreshing archives done.' in main.progressText.text()
    assert tab.bCheck.isEnabled()


def test_repo_prune(qapp, qtbot, archive_env):
    """Test for archive pruning"""
    main, tab = archive_env
    qtbot.mouseClick(tab.bPrune, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: 'Pruning old archives' in main.progressText.text(), **pytest._wait_defaults)
    qtbot.waitUntil(lambda: 'Refreshing archives done.' in main.progressText.text(), **pytest._wait_defaults)


@pytest.mark.min_borg_version('1.2.0a1')
def test_repo_compact(qapp, qtbot, archive_env):
    """Test for archive compaction"""
    main, tab = archive_env
    qtbot.waitUntil(lambda: tab.compactButton.isEnabled(), **pytest._wait_defaults)
    assert tab.compactButton.isEnabled()

    qtbot.mouseClick(tab.compactButton, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: 'compaction freed about' in main.logText.text().lower(), **pytest._wait_defaults)


def test_check(qapp, qtbot, archive_env):
    """Test for archive consistency check"""
    main, tab = archive_env

    qapp.check_failed_event.disconnect()

    qtbot.waitUntil(lambda: tab.bCheck.isEnabled(), **pytest._wait_defaults)
    qtbot.mouseClick(tab.bCheck, QtCore.Qt.MouseButton.LeftButton)
    success_text = 'INFO: Archive consistency check complete'

    qtbot.waitUntil(lambda: success_text in main.logText.text(), **pytest._wait_defaults)


@pytest.mark.skipif(sys.platform == 'darwin', reason="Macos fuse support is uncertain")
def test_mount(qapp, qtbot, monkeypatch, choose_file_dialog, tmpdir, archive_env):
    """Test for archive mounting and unmounting"""

    def psutil_disk_partitions(**kwargs):
        DiskPartitions = namedtuple('DiskPartitions', ['device', 'mountpoint'])
        return [DiskPartitions('borgfs', str(tmpdir))]

    monkeypatch.setattr(psutil, "disk_partitions", psutil_disk_partitions)
    monkeypatch.setattr(vorta.views.archive_tab, "choose_file_dialog", choose_file_dialog)

    main, tab = archive_env
    tab.archiveTable.selectRow(0)

    qtbot.waitUntil(lambda: tab.bMountRepo.isEnabled(), **pytest._wait_defaults)

    qtbot.mouseClick(tab.bMountArchive, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Mounted'), **pytest._wait_defaults)

    tab.bmountarchive_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Un-mounted successfully.'), **pytest._wait_defaults)

    tab.bmountrepo_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Mounted'), **pytest._wait_defaults)

    tab.bmountrepo_clicked()
    qtbot.waitUntil(lambda: tab.mountErrors.text().startswith('Un-mounted successfully.'), **pytest._wait_defaults)


def test_archive_extract(qapp, qtbot, monkeypatch, choose_file_dialog, tmpdir, archive_env):
    """Test for archive extraction"""
    main, tab = archive_env

    tab.archiveTable.selectRow(2)
    tab.extract_action()

    qtbot.waitUntil(lambda: hasattr(tab, '_window'), **pytest._wait_defaults)

    # Select all files
    tree_view = tab._window.treeView.model()
    tree_view.setData(tree_view.index(0, 0), QtCore.Qt.CheckState.Checked, QtCore.Qt.ItemDataRole.CheckStateRole)
    monkeypatch.setattr(vorta.views.archive_tab, "choose_file_dialog", choose_file_dialog)
    qtbot.mouseClick(tab._window.extractButton, QtCore.Qt.MouseButton.LeftButton)

    qtbot.waitUntil(lambda: 'Restored files from archive.' in main.progressText.text(), **pytest._wait_defaults)

    assert [item.basename for item in tmpdir.listdir()] == ['private' if sys.platform == 'darwin' else 'tmp']


def test_archive_delete(qapp, qtbot, mocker, archive_env):
    """Test for archive deletion"""
    main, tab = archive_env

    archivesCount = tab.archiveTable.rowCount()

    mocker.patch.object(vorta.views.archive_tab.ArchiveTab, 'confirm_dialog', lambda x, y, z: True)

    tab.archiveTable.selectRow(0)
    tab.delete_action()
    qtbot.waitUntil(lambda: 'Archive deleted.' in main.progressText.text(), **pytest._wait_defaults)

    assert ArchiveModel.select().count() == archivesCount - 1
    assert tab.archiveTable.rowCount() == archivesCount - 1


def test_archive_rename(qapp, qtbot, mocker, archive_env):
    """Test for archive renaming"""
    main, tab = archive_env

    tab.archiveTable.selectRow(0)
    new_archive_name = 'idf89d8f9d8fd98'
    pos = tab.archiveTable.visualRect(tab.archiveTable.model().index(0, 4)).center()
    qtbot.mouseClick(tab.archiveTable.viewport(), QtCore.Qt.MouseButton.LeftButton, pos=pos)
    qtbot.mouseDClick(tab.archiveTable.viewport(), QtCore.Qt.MouseButton.LeftButton, pos=pos)
    qtbot.keyClicks(tab.archiveTable.viewport().focusWidget(), new_archive_name)
    qtbot.keyClick(tab.archiveTable.viewport().focusWidget(), QtCore.Qt.Key.Key_Return)

    # Successful rename case
    qtbot.waitUntil(lambda: tab.archiveTable.model().index(0, 4).data() == new_archive_name, **pytest._wait_defaults)



import os
import re
import sys

import nox

borg_version = os.getenv("BORG_VERSION")

if borg_version:
    # Use specified borg version
    supported_borgbackup_versions = [borg_version]
else:
    # Default to latest stable version
    supported_borgbackup_versions = ["1.4.3"]


@nox.session(venv_backend="uv")
@nox.parametrize("borgbackup", supported_borgbackup_versions)
def run_tests(session, borgbackup):
    # install borgbackup
    if sys.platform == 'darwin':
        # in macOS there's currently no fuse package which works with borgbackup directly
        session.install(f"borgbackup=={borgbackup}")
    elif borgbackup == "1.1.18":
        # borgbackup 1.1.18 doesn't support pyfuse3
        session.install("llfuse")
        session.install(f"borgbackup[llfuse]=={borgbackup}")
    else:
        session.install(f"borgbackup[pyfuse3]=={borgbackup}")

    # install dev dependencies and package
    session.install("-e", ".[test]")

    # check versions
    cli_version = session.run("borg", "--version", silent=True).strip()
    cli_version = re.search(r"borg (\S+)", cli_version).group(1)
    python_version = session.run("python", "-c", "import borg; print(borg.__version__)", silent=True).strip()

    session.log(f"Borg CLI version: {cli_version}")
    session.log(f"Borg Python version: {python_version}")

    assert cli_version == borgbackup
    assert python_version == borgbackup

    session.run("pytest", *session.posargs, env={"BORG_VERSION": borgbackup})



import os
import subprocess

import pytest
from packaging.version import Version
from peewee import SqliteDatabase
from PyQt6.QtCore import QCoreApplication

import vorta
import vorta.application
import vorta.borg.jobs_manager
from vorta.store.models import (
    ArchiveModel,
    BackupProfileModel,
    EventLogModel,
    RepoModel,
    RepoPassword,
    SchemaVersion,
    SettingsModel,
    SourceFileModel,
    WifiSettingModel,
)
from vorta.utils import borg_compat
from vorta.views.main_window import ArchiveTab, MainWindow


def disconnect_all(signal):
    """
    Disconnect ALL handlers from a Qt signal.
    Unlike signal.disconnect() without arguments which only disconnects ONE handler,
    this function disconnects all connected handlers by calling disconnect in a loop
    until TypeError is raised (indicating no more handlers are connected).
    """
    while True:
        try:
            signal.disconnect()
        except TypeError:
            # No more handlers connected
            break


def all_workers_finished(jobs_manager):
    """
    Check if all worker threads have actually exited.
    This is more thorough than is_worker_running() which only checks current_job,
    because threads may still be alive briefly after current_job is set to None.
    """
    for worker in jobs_manager.workers.values():
        if worker.is_alive():
            return False
    return True


models = [
    RepoModel,
    RepoPassword,
    BackupProfileModel,
    SourceFileModel,
    SettingsModel,
    ArchiveModel,
    WifiSettingModel,
    EventLogModel,
    SchemaVersion,
]


@pytest.fixture(scope='function', autouse=True)
def borg_version():
    borg_version = os.getenv('BORG_VERSION')
    if not borg_version:
        borg_version = subprocess.run(['borg', '--version'], stdout=subprocess.PIPE).stdout.decode('utf-8')
        borg_version = borg_version.split(' ')[1]

    # test window does not automatically set borg version
    borg_compat.set_version(borg_version, borg_compat.path)

    parsed_borg_version = Version(borg_version)
    return borg_version, parsed_borg_version


@pytest.fixture(scope='function', autouse=True)
def create_test_repo(tmpdir_factory, borg_version):
    repo_path = tmpdir_factory.mktemp('repo')
    source_files_dir = tmpdir_factory.mktemp('borg_src')

    is_borg_v2 = borg_version[1] >= Version('2.0.0b1')

    if is_borg_v2:
        subprocess.run(['borg', '-r', str(repo_path), 'rcreate', '--encryption=none'], check=True)
    else:
        subprocess.run(['borg', 'init', '--encryption=none', str(repo_path)], check=True)

    def create_archive(timestamp, name):
        if is_borg_v2:
            subprocess.run(
                ['borg', '-r', str(repo_path), 'create', '--timestamp', timestamp, name, str(source_files_dir)],
                cwd=str(repo_path),
                check=True,
            )
        else:
            subprocess.run(
                ['borg', 'create', '--timestamp', timestamp, f'{repo_path}::{name}', str(source_files_dir)],
                cwd=str(repo_path),
                check=True,
            )

    # /src/file
    file_path = os.path.join(source_files_dir, 'file')
    with open(file_path, 'w') as f:
        f.write('test')

    # /src/dir/
    dir_path = os.path.join(source_files_dir, 'dir')
    os.mkdir(dir_path)

    # /src/dir/file
    file_path = os.path.join(dir_path, 'file')
    with open(file_path, 'w') as f:
        f.write('test')

    # Create first archive
    create_archive('2023-06-14T01:00:00', 'test-archive1')

    # /src/dir/symlink
    symlink_path = os.path.join(dir_path, 'symlink')
    os.symlink(file_path, symlink_path)

    # /src/dir/hardlink
    hardlink_path = os.path.join(dir_path, 'hardlink')
    os.link(file_path, hardlink_path)

    # /src/dir/fifo
    fifo_path = os.path.join(dir_path, 'fifo')
    os.mkfifo(fifo_path)

    # /src/dir/chrdev
    supports_chrdev = True
    try:
        chrdev_path = os.path.join(dir_path, 'chrdev')
        os.mknod(chrdev_path, mode=0o600 | 0o020000)
    except PermissionError:
        supports_chrdev = False

    create_archive('2023-06-14T02:00:00', 'test-archive2')

    # Rename dir to dir1
    os.rename(dir_path, os.path.join(source_files_dir, 'dir1'))

    create_archive('2023-06-14T03:00:00', 'test-archive3')

    # Rename all files under dir1
    for file in os.listdir(os.path.join(source_files_dir, 'dir1')):
        os.rename(os.path.join(source_files_dir, 'dir1', file), os.path.join(source_files_dir, 'dir1', file + '1'))

    create_archive('2023-06-14T04:00:00', 'test-archive4')

    # Delete all file under dir1
    for file in os.listdir(os.path.join(source_files_dir, 'dir1')):
        os.remove(os.path.join(source_files_dir, 'dir1', file))

    create_archive('2023-06-14T05:00:00', 'test-archive5')

    # change permission of dir1
    os.chmod(os.path.join(source_files_dir, 'dir1'), 0o700)

    create_archive('2023-06-14T06:00:00', 'test-archive6')

    return repo_path, source_files_dir, supports_chrdev


@pytest.fixture(scope='function', autouse=True)
def init_db(qapp, qtbot, tmpdir_factory, create_test_repo):
    tmp_db = tmpdir_factory.mktemp('Vorta').join('settings.sqlite')
    mock_db = SqliteDatabase(
        str(tmp_db),
        pragmas={
            'journal_mode': 'wal',
        },
    )
    vorta.store.connection.init_db(mock_db)

    default_profile = BackupProfileModel(name='Default')
    default_profile.save()

    repo_path, source_dir, _ = create_test_repo

    new_repo = RepoModel(url=repo_path)
    new_repo.encryption = 'none'
    new_repo.save()

    default_profile.repo = new_repo.id
    default_profile.dont_run_on_metered_networks = False
    default_profile.validation_on = False
    default_profile.save()

    source_dir = SourceFileModel(dir=source_dir, repo=new_repo, dir_size=12, dir_files_count=3, path_isdir=True)
    source_dir.save()

    qapp.main_window.deleteLater()
    del qapp.main_window
    qapp.main_window = MainWindow(qapp)  # Re-open main window to apply mock data in UI

    qapp.scheduler.schedule_changed.disconnect()

    yield

    # Teardown: cancel jobs and wait for workers to fully exit
    qapp.jobs_manager.cancel_all_jobs()

    # Wait for all worker threads to actually exit (not just for current_job to be None).
    # This is more thorough than is_worker_running() and prevents thread state leakage.
    qtbot.waitUntil(lambda: all_workers_finished(qapp.jobs_manager), **pytest._wait_defaults)

    # Process any pending events to ensure all queued signals are handled
    QCoreApplication.processEvents()

    # Disconnect signals after events are processed
    disconnect_all(qapp.backup_finished_event)
    disconnect_all(qapp.scheduler.schedule_changed)

    # Clear the workers dict to prevent accumulation of dead thread references
    qapp.jobs_manager.workers.clear()
    qapp.jobs_manager.jobs.clear()

    mock_db.close()


@pytest.fixture
def choose_file_dialog(tmpdir):
    class MockFileDialog:
        def __init__(self, *args, **kwargs):
            self.directory = kwargs.get('directory', None)
            self.subdirectory = kwargs.get('subdirectory', None)

        def open(self, func):
            func()

        def selectedFiles(self):
            if self.subdirectory:
                return [str(tmpdir.join(self.subdirectory))]
            elif self.directory:
                return [str(self.directory)]
            else:
                return [str(tmpdir)]

    return MockFileDialog


@pytest.fixture
def rootdir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture()
def archive_env(qapp, qtbot):
    """
    Common setup for integration tests involving the archive tab.
    """
    main: MainWindow = qapp.main_window
    tab: ArchiveTab = main.archiveTab
    main.tabWidget.setCurrentIndex(3)
    tab.refresh_archive_list()
    qtbot.waitUntil(lambda: tab.archiveTable.rowCount() > 0, **pytest._wait_defaults)
    return main, tab


@pytest.fixture(autouse=True)
def min_borg_version(borg_version, request):
    if request.node.get_closest_marker('min_borg_version'):
        parsed_borg_version = borg_version[1]

        if parsed_borg_version < Version(request.node.get_closest_marker('min_borg_version').args[0]):
            pytest.skip(
                'skipped due to borg version requirement for test: {}'.format(
                    request.node.get_closest_marker('min_borg_version').args[0]
                )
            )


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "min_borg_version(): set minimum required borg version for a test",
    )



import tempfile

from PyQt6 import QtCore
from PyQt6.QtWidgets import QDialogButtonBox, QFileDialog, QMessageBox, QToolTip

from vorta.store.models import BackupProfileModel, SourceFileModel
from vorta.views.export_window import ExportWindow


def test_profile_add_delete(qapp, qtbot, mocker):
    """Tests adding and deleting profiles."""
    main = qapp.main_window

    # add profile and ensure it is created as intended
    main.profile_add_action()
    add_profile_window = main.window
    qtbot.keyClicks(add_profile_window.profileNameField, 'Test Profile')
    save_button = add_profile_window.buttonBox.button(QDialogButtonBox.StandardButton.Save)
    qtbot.mouseClick(save_button, QtCore.Qt.MouseButton.LeftButton)
    assert BackupProfileModel.get_or_none(name='Test Profile') is not None
    assert main.profileSelector.currentItem().text() == 'Test Profile'

    # delete the new profile and ensure it is no longer available.
    mocker.patch.object(QMessageBox, 'question', return_value=QMessageBox.StandardButton.Yes)
    qtbot.mouseClick(main.profileDeleteButton, QtCore.Qt.MouseButton.LeftButton)
    assert BackupProfileModel.get_or_none(name='Test Profile') is None
    assert main.profileSelector.currentItem().text() == 'Default'

    # attempt to delete the last remaining profile
    # see that it cannot be deleted, a warning is displayed, and the profile remains
    warning = mocker.patch.object(QToolTip, 'showText')
    qtbot.mouseClick(main.profileDeleteButton, QtCore.Qt.MouseButton.LeftButton)
    assert "Cannot delete the last profile." in warning.call_args[0][1]
    assert BackupProfileModel.get_or_none(name='Default') is not None
    assert main.profileSelector.currentItem().text() == 'Default'


def test_profile_edit(qapp, qtbot):
    """Tests editing/renaming a profile"""
    main = qapp.main_window

    # click to rename profile, clear the name field, type new profile name
    qtbot.mouseClick(main.profileRenameButton, QtCore.Qt.MouseButton.LeftButton)
    edit_profile_window = main.window
    edit_profile_window.profileNameField.setText("")
    qtbot.keyClicks(edit_profile_window.profileNameField, 'Test Profile')
    save_button = edit_profile_window.buttonBox.button(QDialogButtonBox.StandardButton.Save)
    qtbot.mouseClick(save_button, QtCore.Qt.MouseButton.LeftButton)

    # assert a profile by the old name no longer exists, and the newly named profile does exist and is selected.
    assert BackupProfileModel.get_or_none(name='Default') is None
    assert BackupProfileModel.get_or_none(name='Test Profile') is not None
    assert main.profileSelector.currentItem().text() == 'Test Profile'


def test_profile_import_no_duplicate_sources(qapp, qtbot, mocker):
    """Tests importing an existing profile and choosing to overwrite does not add duplicate sources."""
    main = qapp.main_window

    # Create a new profile and add sources
    profile = BackupProfileModel.create(name='Test Profile')
    sources = ['/path/to/source1', '/path/to/source2']
    for source_path in sources:
        SourceFileModel.create(dir=source_path, path=source_path, profile=profile)

    # Export the profile
    with tempfile.NamedTemporaryFile(suffix=".json", delete=False) as temp_file:
        temp_file_path = temp_file.name

    export_window = ExportWindow(profile)
    mocker.patch.object(QFileDialog, 'getSaveFileName', return_value=(temp_file_path, 'JSON (*.json)'))
    export_window.run()

    # Mock the QMessageBox to return 'Yes' to overwrite the profile
    mocker.patch.object(QMessageBox, 'question', return_value=QMessageBox.StandardButton.Yes)

    # Mock the QFileDialog to return the name of the temporary file
    mocker.patch.object(QFileDialog, 'getOpenFileName', return_value=(temp_file_path, 'JSON (*.json)'))

    # Import the profile and choose to overwrite
    main.profile_import_action()

    # Check that the sources in the profile are the same as before and that there are no duplicates
    imported_sources = [source.dir for source in SourceFileModel.select().where(SourceFileModel.profile == profile)]
    assert set(imported_sources) == set(sources)
    assert len(imported_sources) == len(sources)



import os

import pytest
from PyQt6 import QtCore
from PyQt6.QtWidgets import QMessageBox
from test_constants import TEST_TEMP_DIR


@pytest.fixture()
def source_env(qapp, qtbot):
    """Handles setup and teardown for source tab tests."""
    main = qapp.main_window

    main.show()
    qtbot.waitUntil(main.isVisible, **pytest._wait_defaults)

    main.tabWidget.setCurrentIndex(1)  # activate source tab
    tab = main.sourceTab

    qtbot.waitUntil(tab.isVisible, **pytest._wait_defaults)  # wait for the tab
    qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() >= 0, **pytest._wait_defaults)
    yield main, tab

    qapp.processEvents()  # cleanup


@pytest.mark.skip(reason="prone to failure due to background thread")
def test_source_add_remove(qapp, qtbot, mocker, source_env):
    main, tab = source_env
    qtbot.waitUntil(tab.isVisible, **pytest._wait_defaults)  # visibility check

    mocker.patch.object(QMessageBox, "exec")
    mocker.patch('vorta.filedialog.VortaFileSelector.get_paths', return_value=[os.path.join(TEST_TEMP_DIR, 'test')])
    mocker.patch('os.access', return_value=True)

    initial_count = tab.sourceFilesWidget.rowCount()

    # test add
    tab.source_add()
    qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() > initial_count, **pytest._wait_defaults)

    # test remove
    tab.sourceFilesWidget.selectRow(initial_count)  # Select the new row
    qtbot.mouseClick(tab.removeButton, QtCore.Qt.MouseButton.LeftButton)
    qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() == initial_count, **pytest._wait_defaults)


@pytest.mark.skip(reason="prone to failure due to background thread")
def test_sources_update(qapp, qtbot, mocker, source_env):
    """
    Tests the source update button in the source tab
    """
    main, tab = source_env
    update_path_info_spy = mocker.spy(tab, "update_path_info")

    # test that `update_path_info()` has been called for each source path
    qtbot.mouseClick(tab.updateButton, QtCore.Qt.MouseButton.LeftButton)
    assert tab.sourceFilesWidget.rowCount() == 1
    assert update_path_info_spy.call_count == 1

    # add a new source and reset mock
    mocker.patch('os.access', return_value=True)
    mocker.patch(
        'vorta.filedialog.VortaFileSelector.get_paths',
        return_value=[TEST_TEMP_DIR, os.path.join(TEST_TEMP_DIR, 'another')],
    )
    tab.source_add()
    qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() == 2, **pytest._wait_defaults)
    update_path_info_spy.reset_mock()

    # retest that `update_path_info()` has been called for each source path
    qtbot.mouseClick(tab.updateButton, QtCore.Qt.MouseButton.LeftButton)
    assert tab.sourceFilesWidget.rowCount() == 2
    assert update_path_info_spy.call_count == 2


@pytest.mark.skip(reason="prone to failure due to background thread")
def test_source_copy(qapp, qtbot, mocker, source_env):
    main, tab = source_env
    qtbot.waitUntil(tab.isVisible, **pytest._wait_defaults)

    mock_clipboard = mocker.patch.object(qapp.clipboard(), "setMimeData")
    mocker.patch('os.access', return_value=True)
    mocker.patch('vorta.filedialog.VortaFileSelector.get_paths', return_value=[TEST_TEMP_DIR])

    initial_count = tab.sourceFilesWidget.rowCount()
    tab.source_add()
    qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() > initial_count, **pytest._wait_defaults)

    tab.sourceFilesWidget.selectRow(initial_count)
    tab.source_copy()
    assert mock_clipboard.call_count == 1


# This test is for the paste_text() feature that was removed. Kept here for reference or possible future use.
# @pytest.mark.skip(reason="prone to failure due to background thread")
# @pytest.mark.parametrize(
#     "path, valid",
#     [
#         (__file__, True),  # valid path
#         ("test", False),  # invalid path
#         (f"file://{__file__}", True),  # valid - normal path with prefix that will be stripped
#         (f"file://{__file__}\n{__file__}", True),  # valid - two files separated by new line
#         (f"file://{__file__}{__file__}", False),  # invalid - no new line separating file names
#     ],
# )
# def test_valid_and_invalid_source_paths(qapp, qtbot, mocker, source_env, path, valid):
#     """
#     Valid paths will be added as a source.
#     Invalid paths will trigger an alert and not be added as a source.
#     """
#     main, tab = source_env
#     mock_clipboard = mocker.Mock()
#     mock_clipboard.text.return_value = path

#     mocker.patch.object(vorta.views.source_tab.QApplication, 'clipboard', return_value=mock_clipboard)
#     mocker.patch.object(QMessageBox, "exec")  # prevent QMessageBox from stopping test
#     tab.paste_text()

#     if valid:
#         assert not hasattr(tab, '_msg')
#         qtbot.waitUntil(lambda: tab.sourceFilesWidget.rowCount() == 2, **pytest._wait_defaults)
#         assert tab.sourceFilesWidget.rowCount() == 2
#     else:
#         qtbot.waitUntil(lambda: hasattr(tab, "_msg"), **pytest._wait_defaults)
#         assert tab._msg.text().startswith("Some of your sources are invalid")
#         assert tab.sourceFilesWidget.rowCount() == 1
